# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: UTF-8 -*-
"""
Hillas-style moment-based shower image parametrization.
"""

from collections import namedtuple
import numpy as np
from astropy.units import Quantity
from astropy.coordinates import Angle
import astropy.units as u

__all__ = [
    'MomentParameters',
    'hillas_parameters',
    'hillas_parameters_1',
    'hillas_parameters_2',
    'hillas_parameters_3',
    'hillas_parameters_4',
    'HillasParameterizationError',
]

MomentParameters = namedtuple(
    "MomentParameters",
    "size,cen_x,cen_y,length,width,r,phi,psi,miss,skewness,kurtosis"
)


class HillasParameterizationError(RuntimeError):
    pass


def hillas_parameters_1(pix_x, pix_y, image, recalculate_pixels=True):
    """Compute Hillas parameters for a given shower image.

    Reference: Appendix of the Whipple Crab paper Weekes et al. (1998) /
    //MP: 1998 paper has no appendix with the parameters, means 1989 paper
    http://adsabs.harvard.edu/abs/1989ApJ...342..379W
    (corrected for some obvious typos)
    //MP: probably better to use Whipple Reynolds et al 1993 paper:
    http://adsabs.harvard.edu/abs/1993ApJ...404..206R

    Parameters
    ----------
    pix_x : array_like
        Pixel x-coordinate
    pix_y : array_like
        Pixel y-coordinate
    image : array_like
        Pixel values corresponding
    recalculate_pixels : Boolean (default True)
        Recalculate the pixel higher multiples (e.g., if pixels move (!) or
        pixel list changes between calls)


    Returns
    -------
    hillas_parameters : `MomentParameters`
    """
    unit = Quantity(pix_x).unit
    pix_x = Quantity(np.asanyarray(pix_x, dtype=np.float64)).value
    pix_y = Quantity(np.asanyarray(pix_y, dtype=np.float64)).value
    image = np.asanyarray(image, dtype=np.float64)
    assert pix_x.shape == image.shape
    assert pix_y.shape == image.shape

    # Compute image moments
    size = np.sum(image)

    # Sanity check1:
    if size == 0:
        raise HillasParameterizationError(("Empty pixels! Cannot"
                                           " calculate image"
                                           " parameters. Exiting..."))

    mean_x = np.sum(image * pix_x) / size
    mean_y = np.sum(image * pix_y) / size

    # Compute major axis line representation y = a * x + b and correlations
    S_xx = np.sum(image * (pix_x - mean_x) ** 2) / size
    S_yy = np.sum(image * (pix_y - mean_y) ** 2) / size
    S_xy = np.sum(image * (pix_x - mean_x) * (pix_y - mean_y)) / size
    S_xxx = np.sum(image * (pix_x - mean_x) ** 3) / size
    S_yyy = np.sum(image * (pix_y - mean_y) ** 3) / size
    S_xyy = np.sum(image * (pix_x - mean_x) * (pix_y - mean_y) ** 2) / size
    S_xxy = np.sum(image * (pix_y - mean_y) * (pix_x - mean_x) ** 2) / size
    S_x4 = np.sum(image * (pix_x - mean_x) ** 4) / size
    S_y4 = np.sum(image * (pix_y - mean_y) ** 4) / size

    # Sanity check2:

    # If S_xy=0 (which should happen not very often, because Size>0)
    # we cannot calculate Length and Width.  In reallity it is almost
    # impossible to have a distribution of cerenkov photons in the
    # used pixels which is exactly symmetric along one of the axis
    if S_xy == 0:
        raise HillasParameterizationError(("X and Y uncorrelated. Cannot "
                                           "calculate length & width"))

    d0 = S_yy - S_xx
    d1 = 2 * S_xy
    d2 = d0 + np.sqrt(d0 * d0 + d1 * d1)
    a = d2 / d1
    # Angle between ellipse major ax. and x-axis of camera.
    # Will be used for disp
    delta = np.arctan(a)
    b = mean_y - a * mean_x
    # Sin & Cos Will be used for calculating higher order image parameters
    cos_delta = 1 / np.sqrt(1 + a * a)
    sin_delta = a * cos_delta

    # Compute Hillas parameters
    width_2 = (S_yy + a * a * S_xx - 2 * a * S_xy) / \
        (1 + a * a)  # width squared
    length_2 = (S_xx + a * a * S_yy + 2 * a * S_xy) / \
        (1 + a * a)  # length squared

    # Sanity check3:
    width = 0. if width_2 < 0. else np.sqrt(width_2)
    length = 0. if length_2 < 0. else np.sqrt(length_2)

    miss = np.abs(b / np.sqrt(1 + a * a))
    r = np.sqrt(mean_x * mean_x + mean_y * mean_y)
    phi = np.arctan2(mean_y, mean_x)

    # Higher order moments
    sk = cos_delta * (pix_x - mean_x) + sin_delta * (pix_y - mean_y)

    skewness = ((np.sum(image * np.power(sk, 3)) / size) /
                ((np.sum(image * np.power(sk, 2)) / size) ** (3. / 2)))
    kurtosis = ((np.sum(image * np.power(sk, 4)) / size) /
                ((np.sum(image * np.power(sk, 2)) / size) ** 2))
    asym3 = (np.power(cos_delta, 3) * S_xxx
             + 3.0 * np.power(cos_delta, 2) * sin_delta * S_xxy
             + 3.0 * cos_delta * np.power(sin_delta, 2) * S_xyy
             + np.power(sin_delta, 3) * S_yyy)
    asym = - np.power(-asym3, 1. / 3) if (asym3 < 0.) \
        else np.power(asym3, 1. / 3)

    assert np.sign(skewness) == np.sign(asym)

    # another definition of assymetry
    # asym = (mean_x - pix_x[np.argmax(image)]) * cos_delta
    #      + (mean_y - pix_y[np.argmax(image)]) * sin_delta

    # # Compute azwidth by transforming to (p, q) coordinates
    # sin_theta = mean_y / r
    # cos_theta = mean_x / r
    # q = (mean_x - pix_x) * sin_theta + (pix_y - mean_y) * cos_theta
    # m_q = np.sum(image * q) / size
    # m_qq = np.sum(image * q * q) / size
    # azwidth_2 = m_qq - m_q * m_q
    # azwidth = np.sqrt(azwidth_2)

    return MomentParameters(size=size,
                            cen_x=mean_x*unit,
                            cen_y=mean_y*unit,
                            length=length*unit,
                            width=width*unit,
                            r=r*unit,
                            phi=Angle(phi*u.rad),
                            psi=Angle(delta*u.rad),
                            miss=miss*unit,
                            skewness=skewness,
                            kurtosis=kurtosis)

def static_pix(pix_x, pix_y, recalculate_pixels):
    """Hold static variables for a given camera's pixel positions,
    #if the camera's pixel positions haven't changed since last call, and otherwise or
    if first call initializes them to the right values.

    Parameters
    ----------
    pix_x : array_like
        Pixel x-coordinate
    pix_y : array_like
        Pixel y-coordinate
    recalculate_pixels : Boolean
        Recalculate the pixel higher multiples (e.g., if pixels move (!) or pixel list changes between calls)

    Returns
    -------
    Nothing, but keeps variables as attributes to the function, so acts like a static variable holder."""

    # If not called before or recalculate_pixels
    if (not hasattr(static_pix, "pixdata") or recalculate_pixels): # \
      #or not (np.array_equal(pix_x,static_xy.pix_x) and np.array_equal(pix_y,static_xy.pix_y)):
      #, or if the pixel positions have changed, but this adds 15% calculation time
        static_pix.pixdata = np.row_stack([pix_x,
                                           pix_y,
                                           pix_x * pix_x,
                                           pix_x * pix_y,
                                           pix_y * pix_y])

        static_pix.pixdataHO = \
        np.row_stack([pix_x * static_pix.pixdata[2],
                      static_pix.pixdata[2] * pix_y,
                      pix_x * static_pix.pixdata[4],
                      pix_y * static_pix.pixdata[4],
                      static_pix.pixdata[2] * static_pix.pixdata[2],
                      static_pix.pixdata[2] * static_pix.pixdata[3],
                      static_pix.pixdata[2] * static_pix.pixdata[4],
                      static_pix.pixdata[3] * static_pix.pixdata[4],
                      static_pix.pixdata[4] * static_pix.pixdata[4]])


def hillas_parameters_2(pix_x, pix_y, image, recalculate_pixels=True):
    """Compute Hillas parameters for a given shower image.

    Alternate implementation of `hillas_parameters` ...
    in the end we'll just keep one, but we're using Hillas parameter
    computation as an example for performance checks.

    Parameters
    ----------
    pix_x : array_like
        Pixel x-coordinate
    pix_y : array_like
        Pixel y-coordinate
    image : array_like
        Pixel values corresponding
    recalculate_pixels : Boolean (default True)
        Recalculate the pixel higher multiples (e.g., if pixels move (!) or pixel list changes between calls)

    Returns
    -------
    hillas_parameters : `MomentParameters`
    """

    if type(pix_x)==Quantity:
        unit = pix_x.unit
        assert pix_x.unit == pix_y.unit
    else:
        unit = 1.0

    pix_x = Quantity(np.asanyarray(pix_x, dtype=np.float64)).value
    pix_y = Quantity(np.asanyarray(pix_y, dtype=np.float64)).value
    image = np.asanyarray(image)

    assert pix_x.shape == image.shape
    assert pix_y.shape == image.shape

    size = image.sum()

    if size == 0.0:
        raise (HillasParameterizationError("Empty pixels! Cannot calculate image parameters. Exiting..."))

    #call static_xy to initialize the "static variables"
    #actually, would be nice to just call this if we know the pixel positions have changed
    static_pix(pix_x, pix_y, recalculate_pixels)

    # Compute image moments (done in a bit faster way, but putting all
    # into one 2D array, where each row will be summed to calculate a
    # moment) However, this doesn't avoid a temporary created for the
    # 2D array


    momdata = static_pix.pixdata * image
    moms = momdata.sum(axis=1) / size
    momdataHO = static_pix.pixdataHO * image
    momsHO = momdataHO.sum(axis=1) / size

    # give the moms values comprehensible names
    xm, ym, x2m, xym, y2m = moms
    x3m, x2ym, xy2m, y3m, x4m, x3ym, x2y2m, xy3m, y4m = momsHO

    # intermediate variables (could be avoided if compiler which understands powers, etc)
    xm2 = xm * xm
    ym2 = ym * ym
    xmym = xm * ym

    # calculate variances

    vx2 = x2m - xm2
    vy2 = y2m - ym2
    vxy = xym - xmym
    vx3 = x3m - 3.0 * xm * x2m + 2.0 * xm2 * xm
    vx2y = x2ym - x2m * ym - 2.0 * xym * xm + 2.0 * xm2 * ym
    vxy2 = xy2m - y2m * xm - 2.0 * xym * ym + 2.0 * xm * ym2
    vy3 = y3m - 3.0 * ym * y2m + 2.0 * ym2 * ym

    # polar coordinates of centroid

    rr = np.sqrt(xm2 + ym2)  # could use hypot(xm, ym), but already have squares
    phi = np.arctan2(ym, xm)*u.rad

    # common factors:

    dd = vy2 - vx2
    zz = np.hypot(dd, 2.0 * vxy) # for simpler formulae for length & width suggested CA 901019

    # shower shape parameters

    length = np.sqrt((vx2 + vy2 + zz)/ 2.0)
    width  = np.sqrt((vx2 + vy2 - zz)/2.0)

    # miss, simpler formula for miss introduced CA, 901101; revised MP 910112

    uu = 1.0 + dd / zz
    vv = 2.0 - uu
    miss = np.sqrt((uu * xm2 + vv * ym2) / 2.0
                   - xmym * 2.0 * vxy / zz)
    # maybe for case zz = 0, put miss = dist?

    # rotation angle of ellipse relative to centroid
    tanpsi_numer = (dd + zz) * ym + (2.0 * vxy * xm)
    tanpsi_denom = (2.0 * vxy * ym) - (dd - zz) * xm

    psi = (np.arctan2(tanpsi_numer, tanpsi_denom)) * u.rad

    # -- Asymmetry and other higher moments
    if length != 0.0:
        vx4 = x4m - 4.0 * xm * x3m + 6.0 * xm2 * x2m - 3.0 * xm2 * xm2
        vx3y = x3ym - 3.0 * xm * x2ym  + 3.0 * xm2 * xym - x3m * ym \
                    + 3.0 * x2m * xmym - 3.0 * xm2 * xmym
        vx2y2 = x2y2m - 2.0 * ym * x2ym + x2m * ym2 \
                      - 2.0 * xm * xy2m + 4.0 * xym * xmym + xm2 * y2m - 3.0 * xm2 * ym2
        vxy3 = xy3m - 3.0 * ym * xy2m  + 3.0 * ym2 * xym - y3m * xm \
                    + 3.0 * y2m * xmym - 3.0 * ym2 * xmym
        vy4 = y4m - 4.0 * ym * y3m + 6.0 * ym2 * y2m - 3.0 * ym2 * ym2

        hyp = np.hypot(tanpsi_numer,tanpsi_denom)
        if hyp != 0.:
            cpsi = tanpsi_denom / hyp
            spsi = tanpsi_numer / hyp
        else:
            cpsi = 1.
            spsi = 0.

        cpsi2 = cpsi * cpsi
        spsi2 = spsi * spsi
        cspsi = cpsi * spsi
        sk3bylen3 = (vx3 * cpsi*cpsi2 +
                     3.0 * vx2y * cpsi2 * spsi +
                     3.0 * vxy2 * cpsi * spsi2 +
                     vy3 * spsi*spsi2)
        asym = np.copysign(np.power(np.abs(sk3bylen3),1./3.), sk3bylen3) / length
        skewness = asym*asym*asym  # for MP's asym... (not for MAGIC asym!)
        # Kurtosis
        kurt = (vx4 * cpsi2 * cpsi2 +
                    4.0 * vx3y * cpsi2 * cspsi +
                    6.0 * vx2y2 * cpsi2 * spsi2 +
                    4.0 * vxy3 * cspsi * spsi2 +
                    vy4 * spsi2 * spsi2)
        kurtosis = kurt / (length*length*length*length)
    else:  # Skip Higher Moments
        psi = 0.0*u.rad
        skewness = 0.0
        kurtosis = 0.0
        asym = 0.0

    return MomentParameters(size=size, cen_x=xm*unit, cen_y=ym*unit,
                            length=length*unit, width=width*unit, r=rr*unit,
                            phi=Angle(phi), psi=Angle(psi),
                            miss=miss*unit, skewness=skewness,
                            kurtosis=kurtosis)


def hillas_parameters_3(pix_x, pix_y, image, recalculate_pixels=True):
    """Compute Hillas parameters for a given shower image.

    MP: probably better to use Whipple Reynolds et al 1993 paper:
    http://adsabs.harvard.edu/abs/1993ApJ...404..206R
    which should be the same as one of my ICRC 1991 papers and my thesis.

    Parameters
    ----------
    pix_x : array_like
        Pixel x-coordinate
    pix_y : array_like
        Pixel y-coordinate
    image : array_like
        Pixel values corresponding
    recalculate_pixels : Boolean (default True)
        Recalculate the pixel higher multiples (e.g., if pixels move (!) or pixel list changes between calls)

    Returns
    -------
    hillas_parameters : `MomentParameters`
    """

    if type(pix_x)==Quantity:
        unit = pix_x.unit
        assert pix_x.unit == pix_y.unit
    else:
        unit = 1.0

    pix_x = Quantity(np.asanyarray(pix_x, dtype=np.float64)).value
    pix_y = Quantity(np.asanyarray(pix_y, dtype=np.float64)).value

    # make sure they are numpy arrays so we can use numpy operations
    #pix_x = np.asanyarray(pix_x)
    #pix_y = np.asanyarray(pix_y)
    image = np.asanyarray(image, dtype=np.float64)

    assert pix_x.shape == image.shape
    assert pix_y.shape == image.shape

    # Code to interface with historic version
    wxdeg = pix_x
    wydeg = pix_y
    event = image

    # Code below is from the historical Whipple routine, with original comments
    # MJL is Mark Lang, MP is Michael Punch, GV is Giuseppe Vacanti, CA is Carl Akerlof, RCL is Dick Lamb,

    # Translated from ForTran to c with "fable": https://github.com/ctessum/fable-go
    # and from c to Python with "cpp2python:  https://github.com/andreikop/cpp2python , then edited by hand
    # C***********************************************************************
    # C                              HILLAKPAR                               *
    # C***********************************************************************
    # C-- This version also calculates the asymmetry of the image.
    # C-- Use of Akerlof Azwidth (Akwidth?) implemented MP 910112
    # C-- Simpler Miss formula suggested CA, MP, 901101
    # C-- Simpler Width and Length formulae suggested CA, 901017
    # C-- Yet another little bug fixed by MP 900418- Generalize the case of
    # C--      the horizontal line.
    # C-- Bug fixed by RCL 900307-The case of the horizontal line image was
    # C--     never considered.
    # C-- Bug fixed by GV 900215
    # C** This version takes events in WHIPPLE format and parameterises them.
    # C**    M. Punch ,900105
    # C-- G. Vacanti introduces the common statement: coordinates are
    # C-- computed only in the main program.
    # C-- Modified by M. Punch to make it faster April, 1989
    # C-- mjl 10 dec 87
    # C--
    # -- routine to calculate the six hillas iarameters

    sumsig , sumxsig, sumysig, sumx2sig, sumy2sig, sumxysig, sumx3sig, sumx2ysig, sumxy2sig, sumy3sig = np.zeros(10)

    for i in range(np.size(event)):
        if event[i] != 0.0:
            wxbyev = wxdeg[i] * event[i]
            wybyev = wydeg[i] * event[i]
            sumsig += event[i]
            sumxsig += wxbyev
            sumx2sig += wxdeg[i] * wxbyev
            sumysig += wybyev
            sumy2sig += wydeg[i] * wybyev
            sumxysig += wxdeg[i] * wybyev
            sumx3sig += wxdeg[i] * wxdeg[i] * wxbyev
            sumx2ysig += wxdeg[i] * wxdeg[i] * wybyev
            sumxy2sig += wxdeg[i] * wydeg[i] * wybyev
            sumy3sig += wydeg[i] * wydeg[i] * wybyev

    if sumsig == 0.0:
        raise (HillasParameterizationError("Empty pixels! Cannot calculate image parameters. Exiting..."))

    xm = sumxsig / sumsig
    x2m = sumx2sig / sumsig
    ym = sumysig / sumsig
    y2m = sumy2sig / sumsig
    xym = sumxysig / sumsig

    xm2 = xm * xm
    ym2 = ym * ym
    xmym = xm * ym

    vx2 = x2m - xm2
    vy2 = y2m - ym2
    vxy = xym - xmym

    x3m = sumx3sig / sumsig
    x2ym = sumx2ysig / sumsig
    xy2m = sumxy2sig / sumsig
    y3m = sumy3sig / sumsig

    vx3 = x3m - 3.0 * xm * x2m + 2.0 * xm2 * xm
    vx2y = x2ym - x2m * ym - 2.0 * xym * xm + 2.0 * xm2 * ym
    vxy2 = xy2m - y2m * xm - 2.0 * xym * ym + 2.0 * xm * ym2
    vy3 = y3m - 3.0 * ym * y2m + 2.0 * ym2 * ym

    d = vy2 - vx2
    dist = np.sqrt(xm2 + ym2)
    phi = np.arctan2(ym, xm)

    # -- simpler formulae for length & width suggested CA 901019
    z = np.sqrt(d * d + 4.0 * vxy * vxy)
    length = np.sqrt((vx2 + vy2 + z) / 2.0)
    width = np.sqrt((vy2 + vx2 - z) / 2.0)

    # -- simpler formula for miss introduced CA, 901101
    # -- revised MP 910112
    if z == 0.0:
        miss = dist
    else:
        uu = 1 + d / z
        vv = 2 - uu
        miss = np.sqrt((uu * xm2 + vv * ym2) / 2.0 - xmym * (2.0 * vxy / z))

    # Code to de-interface with historical code
    size = sumsig
    m_x = xm
    m_y = ym
    r = dist

    psi = np.arctan2((d + z) * ym + 2.0 * vxy * xm, 2.0 * vxy * ym - (d - z) * xm)
    cpsi = np.cos(psi)
    spsi = np.sin(psi)


    # -- Asymmetry
    if length == 0.0:
        asymm = 0.0
    else:
        asymm = (vx3 * np.power(cpsi,3) +
                 3.0 * vx2y * spsi * np.power(cpsi,2) + 3.0 * vxy2 * cpsi * np.power(spsi,2) +
                 vy3 * np.power(spsi,3))
        asymm = np.copysign(np.exp(np.log(np.abs(asymm)) / 3.0), asymm) / length

    # # -- Akerlof azwidth now used, 910112
    # d = y2m - x2m
    # z = np.sqrt(d * d + 4 * xym * xym)
    # azwidth = np.sqrt((x2m + y2m - z) / 2.0)
    #
    # isize = int(sumsig)

    # Code to de-interface with historical code
    skewness = asymm*asymm*asymm
    kurtosis = np.nan

    return MomentParameters(size=size, cen_x=m_x*unit, cen_y=m_y*unit,
                            length=length*unit, width=width*unit,
                            r=r*unit, phi=Angle(phi*u.rad),
                            psi=Angle(psi*u.rad),
                            miss=miss*unit,
                            skewness=skewness, kurtosis=kurtosis)

def static_xy(pix_x, pix_y, recalculate_pixels):
    """Hold static variables for a given camera's pixel positions,
    #if the camera's pixel positions haven't changed since last call, and otherwise or
    if first call initializes them to the right values.

    Parameters
    ----------
    pix_x : array_like
        Pixel x-coordinate
    pix_y : array_like
        Pixel y-coordinate
    recalculate_pixels : Boolean
        Recalculate the pixel higher multiples (e.g., if pixels move (!) or pixel list changes between calls)

    Returns
    -------
    Nothing, but keeps variables as attributes to the function, so acts like a static variable holder."""

    # If not called before or recalculate_pixels
    if (not hasattr(static_xy, "pix_x") or recalculate_pixels): # \
      #or not (np.array_equal(pix_x,static_xy.pix_x) and np.array_equal(pix_y,static_xy.pix_y)):
      #, or if the pixel positions have changed, but this adds 15% calculation time
        static_xy.pix_x = pix_x
        static_xy.pix_y = pix_y
        static_xy.pix_x2 = pix_x * pix_x
        static_xy.pix_y2 = pix_y * pix_y
        static_xy.pix_xy = pix_x * pix_y
        static_xy.pix_x3 = static_xy.pix_x2 * pix_x
        static_xy.pix_x2y = static_xy.pix_x2 * pix_y
        static_xy.pix_xy2 = pix_x * static_xy.pix_y2

        static_xy.pix_y3 = pix_y * static_xy.pix_y2
        static_xy.pix_x4 = static_xy.pix_x3 * pix_x
        static_xy.pix_x3y = static_xy.pix_x3 * pix_y
        static_xy.pix_x2y2 = static_xy.pix_x2 * static_xy.pix_y2
        static_xy.pix_xy3 = pix_x * static_xy.pix_y3
        static_xy.pix_y4 = static_xy.pix_y3 * pix_y


def hillas_parameters_4(pix_x, pix_y, image, recalculate_pixels=True):
    """Compute Hillas parameters for a given shower image.

    As for hillas_parameters_3 (old Whipple Fortran code), but more Pythonized

    MP: Parameters calculated as Whipple Reynolds et al 1993 paper:
    http://adsabs.harvard.edu/abs/1993ApJ...404..206R
    which should be the same as one of my ICRC 1991 papers and my thesis.

    Parameters
    ----------
    pix_x : array_like
        Pixel x-coordinate
    pix_y : array_like
        Pixel y-coordinate
    image : array_like
        Pixel values corresponding
    recalculate_pixels : Boolean (default True)
        Recalculate the pixel higher multiples (e.g., if pixels move (!) or pixel list changes between calls)

    Returns
    -------
    hillas_parameters : `MomentParameters`
    """


    if type(pix_x)==Quantity:
        unit = pix_x.unit
        assert pix_x.unit == pix_y.unit
    else:
        unit = 1.0
    # MP: Actually, I don't know why we need to strip the units... shouldn' the calculations all work with them?

    pix_x = Quantity(np.asanyarray(pix_x, dtype=np.float64)).value
    pix_y = Quantity(np.asanyarray(pix_y, dtype=np.float64)).value
    image = np.asanyarray(image, dtype=np.float64)
    assert pix_x.shape == image.shape
    assert pix_y.shape == image.shape

    (sumsig , sumxsig, sumysig, sumx2sig, sumy2sig, sumxysig, sumx3sig,
     sumx2ysig, sumxy2sig, sumy3sig) = np.zeros(10)

    # Call static_xy to initialize the "static variables"
    # Actually, would be nice to just call this if we know the pixel positions have changed
    static_xy(pix_x, pix_y, recalculate_pixels)

    sumsig = image.sum()
    sumxsig = (image * pix_x).sum()
    sumysig = (image * pix_y).sum()
    sumx2sig = (image * static_xy.pix_x2).sum()
    sumy2sig = (image * static_xy.pix_y2).sum()
    sumxysig = (image * static_xy.pix_xy).sum()


    sumx3sig = (image * static_xy.pix_x3).sum()
    sumx2ysig = (image * static_xy.pix_x2y).sum()
    sumxy2sig = (image * static_xy.pix_xy2).sum()
    sumy3sig = (image * static_xy.pix_y3).sum()

    sumx4sig = (image * static_xy.pix_x4).sum()
    sumx3ysig = (image * static_xy.pix_x3y).sum()
    sumx2y2sig = (image * static_xy.pix_x2y2).sum()
    sumxy3sig = (image * static_xy.pix_xy3).sum()
    sumy4sig = (image * static_xy.pix_y4).sum()

    if sumsig == 0.0:
        raise (HillasParameterizationError("Empty pixels! Cannot calculate image parameters. Exiting..."))

    xm   = sumxsig   / sumsig
    ym   = sumysig   / sumsig
    x2m  = sumx2sig  / sumsig
    y2m  = sumy2sig  / sumsig
    xym  = sumxysig  / sumsig


    x3m  = sumx3sig  / sumsig
    x2ym = sumx2ysig / sumsig
    xy2m = sumxy2sig / sumsig
    y3m  = sumy3sig  / sumsig

    x4m  = sumx4sig  / sumsig
    x3ym = sumx3ysig / sumsig
    x2y2m = sumx2y2sig / sumsig
    xy3m = sumxy3sig / sumsig
    y4m  = sumy4sig  / sumsig

    # Doing this should be same as above, but its 4us slower !?
    #(xm, ym, x2m, y2m, xym, x3m, x2ym, xy2m, y3m) = \
    #    (sumxsig, sumysig, sumx2sig, sumy2sig, sumxysig, sumx3sig,
    # sumx2ysig, sumxy2sig, sumy3sig) / sumsig

    xm2 = xm * xm
    ym2 = ym * ym
    xmym = xm * ym

    vx2 = x2m - xm2
    vy2 = y2m - ym2
    vxy = xym - xmym

    vx3 = x3m - 3.0 * xm * x2m + 2.0 * xm2 * xm
    vx2y = x2ym - x2m * ym - 2.0 * xym * xm + 2.0 * xm2 * ym
    vxy2 = xy2m - y2m * xm - 2.0 * xym * ym + 2.0 * xm * ym2
    vy3 = y3m - 3.0 * ym * y2m + 2.0 * ym2 * ym

    d = vy2 - vx2
    dist = np.sqrt(xm2 + ym2) #could use hypot(xm,ym), but already have squares
    phi = np.arctan2(ym, xm)

    # -- simpler formulae for length & width suggested CA 901019
    z = np.hypot(d, 2.0 * vxy)
    length = np.sqrt((vx2 + vy2 + z) / 2.0)
    width = np.sqrt((vy2 + vx2 - z) / 2.0)

    # -- simpler formula for miss introduced CA, 901101
    # -- revised MP 910112
    if z == 0.0:
        miss = dist
    else:
        uu = 1 + d / z
        vv = 2 - uu
        miss = np.sqrt((uu * xm2 + vv * ym2) / 2.0 - xmym * (2.0 * vxy / z))

    #Change to faster caluclation of psi and avoid inaccuracy for hyp
    #psi = np.arctan2((d + z) * ym + 2.0 * vxy * xm, 2.0 *vxy * ym - (d - z) * xm)
    #hyp = np.sqrt(2 * z * (z + d))  #! should be simplification of sqrt((d+z)**2+(2*vxy)**2 ... but not accurate!
    #hyp = np.hypot(d + z,2 * vxy)
    #psi = np.arctan2(d + z, 2 * vxy)
    #cpsi = np.cos(psi)
    #spsi = np.sin(psi)
    tanpsi_numer = (d + z) * ym + 2.0 * vxy * xm
    tanpsi_denom = 2.0 *vxy * ym - (d - z) * xm
    psi = np.arctan2(tanpsi_numer,tanpsi_denom)

    # Code to de-interface with historical code
    size = sumsig
    m_x = xm
    m_y = ym
    length = length
    r = dist


    # Note, "skewness" is the same as the Whipple/MP "asymmetry^3", which is fine.
    # ... and also, Whipple/MP "asymmetry" * "length" = MAGIC "asymmetry"
    # ... so, MAGIC "asymmetry" = MAGIC "skewness"^(1/3) * "length"
    # I don't know what MAGIC's "asymmetry" is supposed to be.

    # -- Asymmetry and other higher moments
    if length != 0.0:
        vx4 = x4m - 4.0 * xm * x3m + 6.0 * xm2 * x2m - 3.0 * xm2 * xm2
        vx3y = x3ym - 3.0 * xm * x2ym + 3.0 * xm2 * xym - x3m * ym \
                  + 3.0 * x2m * xmym - 3.0 * xm2 * xm * ym
        vx2y2 = x2y2m - 2.0 * ym * x2ym + x2m * ym2 \
                   - 2.0 * xm * xy2m + 4.0 * xym * xmym + xm2 * y2m - 3.0 * xm2 * ym2
        vxy3 = xy3m - 3.0 * ym * xy2m + 3.0 * ym2 * xym - y3m * xm \
                  + 3.0 * y2m * xmym - 3.0 * ym2 * ym * xm
        vy4 = y4m - 4.0 * ym * y3m + 6.0 * ym2 * y2m - 3.0 * ym2 * ym2
        
        hyp = np.hypot(tanpsi_numer, tanpsi_denom)
        if hyp != 0.:
            cpsi = tanpsi_denom / hyp
            spsi = tanpsi_numer / hyp
        else:
            cpsi = 1.
            spsi = 0.

        cpsi2 = cpsi * cpsi
        spsi2 = spsi * spsi
        cspsi = cpsi * spsi

        sk3bylen3 = (vx3 * cpsi*cpsi2 +
                    3.0 * vx2y * cpsi2 * spsi +
                    3.0 * vxy2 * cpsi * spsi2 +
                    vy3 * spsi*spsi2)
        asym = np.copysign(np.power(np.abs(sk3bylen3),1./3.), sk3bylen3) / length
        skewness = asym*asym*asym  # for MP's asym... (not for MAGIC asym!)

        # Kurtosis
        kurt = (vx4 * cpsi2 * cpsi2 +
                    4.0 * vx3y * cpsi2 * cspsi +
                    6.0 * vx2y2 * cpsi2 * spsi2 +
                    4.0 * vxy3 * cspsi * spsi2 +
                    vy4 * spsi2 * spsi2)
        kurtosis = kurt / (length*length*length*length)

    else:  # Skip Higher Moments
        asym = 0.0
        psi = 0.0
        skewness = 0.0
        kurtosis = 0.0

    # Azwidth not used anymore
    # # -- Akerlof azwidth now used, 910112
    # d = y2m - x2m
    # z = np.sqrt(d * d + 4 * xym * xym)
    # azwidth = np.sqrt((x2m + y2m - z) / 2.0)

    return MomentParameters(size=size, cen_x=m_x*unit, cen_y=m_y*unit,
                            length=length*unit, width=width*unit, r=r*unit,
                            phi=Angle(phi*u.rad),
                            psi=Angle(psi*u.rad),
                            miss=miss*unit,
                            skewness=skewness, kurtosis=kurtosis)

# use the 4 version by default.
hillas_parameters = hillas_parameters_4




from ctapipe.instrument import CameraGeometry
from ctapipe.image import tailcuts_clean, toymodel
from ctapipe.image.hillas import (hillas_parameters_1, hillas_parameters_2,
                                  hillas_parameters_3, hillas_parameters_4)
from astropy import units as u
from numpy import isclose
from numpy.random import seed

def create_sample_image(psi='-30d'):

    seed(10)

    # set up the sample image using a HESS camera geometry (since it's easy
    # to load)
    geom = CameraGeometry.from_name("LSTCam")

    # make a toymodel shower model
    model = toymodel.generate_2d_shower_model(centroid=(0.2, 0.3),
                                              width=0.001, length=0.01,
                                              psi=psi)

    # generate toymodel image in camera for this shower model.
    image, signal, noise = toymodel.make_toymodel_shower_image(geom, model.pdf,
                                                               intensity=50,
                                                               nsb_level_pe=100)

    # denoise the image, so we can calculate hillas params
    clean_mask = tailcuts_clean(geom, image, 10,
                                5)  # pedvars = 1 and core and boundary
    # threshold in pe
    image[~clean_mask] = 0

    # Pixel values in the camera
    pix_x = geom.pix_x.value
    pix_y = geom.pix_y.value

    return pix_x, pix_y, image

def compare_result(x,y):
    ux = u.Quantity(x)
    uy = u.Quantity(y)
    assert isclose(ux.value,uy.value)
    assert ux.unit == uy.unit


def do_test_hillas(withunits=True):
    """
    test all Hillas-parameter routines on a sample image and see if they
    agree with eachother and with the toy model (assuming the toy model code
    is correct)
    """

    # try all quadrants
    for psi_angle in ['30d','120d','-30d','-120d']:

        px, py, image = create_sample_image(psi_angle)
        results = {}

        if withunits:
            px = px * u.cm
            py = py * u.cm

        results['v1'] = hillas_parameters_1(px, py, image)
        results['v2'] = hillas_parameters_2(px, py, image)
        results['v3'] = hillas_parameters_3(px, py, image)
        results['v4'] = hillas_parameters_4(px, py, image)
        # compare each method's output
        for aa in results:
            for bb in results:
                if aa is not bb:
                    print("comparing {} to {}".format(aa,bb))
                    compare_result(results[aa].length, results[bb].length)
                    compare_result(results[aa].width, results[bb].width)
                    compare_result(results[aa].r, results[bb].r)
                    compare_result(results[aa].phi.deg, results[bb].phi.deg)
                    compare_result(results[aa].psi.deg, results[bb].psi.deg)
                    compare_result(results[aa].miss, results[bb].miss)
                    compare_result(results[aa].skewness, results[bb].skewness)
                    #compare_result(results[aa].kurtosis, results[bb].kurtosis)


def test_hillas_with_units():
    do_test_hillas(withunits=True)

def test_hillas_unitless():
    do_test_hillas(withunits=False)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# coding: utf8
from types import GeneratorType


class StagerSequential():

    """`StagerSequential` class represents a Stager pipeline Step.
    """

    def __init__(self, coroutine, name=None, connections=None, main_connection_name=None):
        """
        Parameters
        ----------
        coroutine : Class instance that contains init, run and finish methods
        connections: list(str)
            define next available steps
        """
        self.name = name
        self.coroutine = coroutine
        self.main_connection_name = main_connection_name
        self.connections = connections or []
        self.running = 0
        self.nb_job_done = 0

    def init(self):
        """
        Initialise coroutine sockets and poller

        Returns
        -------
        True if coroutine init method returns True, otherwise False
        """
        if self.name is None:
            self.name = "STAGER"
        if self.coroutine is None:
            return False
        if self.coroutine.init() == False:
            return False
        self.coroutine.connections = self.connections
        return True

    def run(self, inputs=None):
        """ Executes coroutine run method

        Parameters
        ----------
        inputs: object
             input for coroutine.run
        """
        result = self.coroutine.run(inputs)
        if isinstance(result, GeneratorType):
            for val in result:
                msg, destination = self.get_destination_msg_from_result(val)
                yield (msg, destination)
        else:
            msg, destination = self.get_destination_msg_from_result(result)
            yield (msg, destination)
        self.nb_job_done += 1

    def get_destination_msg_from_result(self, result):
        """If result is a tuple, check if last tuple elem is a valid next
        step.  If yes, return a destination defined to the last tuple
        elem and send result without the destination If no return None
        as destination

        Parameters
        ----------
        result : any type
            value to send (can contain next step name)

        Returns
        -------
        msg, destination

        """
        destination = self.main_connection_name
        if isinstance(result, tuple):
            # look is last tuple elem is a valid next step
            if result[-1] in self.connections.keys():
                destination = result[-1]
                if len(result[:-1]) == 1:
                    msg = result[:-1][0]
                else:
                    msg = result[:-1]
                return msg, destination
            else:
                return result, destination
        else:
            return result, destination

    def finish(self):
        """
        Call coroutine finish method
        """
        self.coroutine.finish()
        return True



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import os
import sys
import logging
from importlib import import_module
from ctapipe.core.tool import Tool



__all__ = ['dynamic_class_from_module', ]


class DynamicClassError(Exception):
    def __init__(self, msg):
        '''Mentions that an exception occurred in the dynamic_class_from_module.
        '''
        self.msg = msg

#def dynamic_class_from_module(class_name, module,  configuration=None):
def dynamic_class_from_module(class_name, module,  tool=None):
    """
    Create an instance of a class from a configuration service section name

    Parameters
    ----------
    module: str
        a python module file name. This module containe class to instantiate
    class_name: str
        python class name contained in module

    Returns
    -------
    A python object instance of a class_name

    Raises
    ------
    """
    if  module == None :
        return None
    logger = logging.getLogger(__name__)
    try:
        _class = getattr(import_module(module), class_name)
        if isinstance(tool,Tool):
            instance = _class(tool,config=tool.config)
        else:
            instance = _class()
        return instance
    except AttributeError as e:
        raise DynamicClassError("Could not create an instance of {} in module {}: {}"
            .format(class_name, module, e))
    except ImportError as e:
        raise DynamicClassError("Could not create an instance of {} in module {}: {}"
            .format(class_name, module, e))
    except TypeError as e:
        raise DynamicClassError("Could not create an instance of {} in module {}: {}"
            .format(class_name, module, e))



from ctapipe.instrument import SubarrayDescription, TelescopeDescription, CameraGeometry
import numpy as np
from astropy import units as u

def test_subarray_description():

    pos = {}
    tel = {}
    foclen = 16 * u.m
    pix_x = np.arange(1764, dtype=np.float) * u.m
    pix_y = np.arange(1764, dtype=np.float) * u.m

    for ii in range(10):

        tel[ii] = TelescopeDescription.guess(pix_x, pix_y, foclen)
        pos[ii] = (np.random.uniform(200, size=2)-100) * u.m



    sub = SubarrayDescription("test array",
                              tel_positions=pos,
                              tel_descriptions=tel)

    sub.info()

    assert sub.num_tels == 10
    assert sub.tel[0].camera is not None
    assert len(sub.to_table()) == 10

    subsub = sub.select_subarray("newsub", [1,2,3,4])
    assert subsub.num_tels == 4
    assert set(subsub.tels.keys()) == {1,2,3,4}


from ctapipe.io.containers import R0CameraContainer, MCEventContainer
from ctapipe.io.hdftableio import HDF5TableWriter, HDF5TableReader
import numpy as np
from astropy import units as u
import tables
import pytest

@pytest.fixture(scope='session')
def temp_h5_file(tmpdir_factory):
    """ a fixture that fetches a temporary output dir/file for a test
    file that we want to read or write (so it doesn't clutter up the test 
    directory when the automated tests are run)"""
    return str(tmpdir_factory.mktemp('data').join('test.h5'))


def test_write_container(temp_h5_file):
    r0tel = R0CameraContainer()
    mc = MCEventContainer()
    mc.reset()
    r0tel.adc_samples = np.random.uniform(size=(50, 10))
    r0tel.adc_sums = np.random.uniform(size=50)
    r0tel.num_samples = 10
    r0tel.meta['test_attribute'] = 3.14159
    r0tel.meta['date'] = "2020-10-10"

    writer = HDF5TableWriter(str(temp_h5_file), group_name='R0',
                             filters=tables.Filters(
        complevel=7))
    writer.exclude("tel_002",".*samples")  # test exclusion of columns

    for ii in range(100):
        r0tel.adc_samples[:] = np.random.uniform(size=(50, 10))
        r0tel.adc_sums[:] = np.random.uniform(size=50)
        r0tel.num_samples = 10
        mc.energy = 10**np.random.uniform(1,2) * u.TeV
        mc.core_x = np.random.uniform(-1, 1) * u.m
        mc.core_y = np.random.uniform(-1, 1) * u.m

        writer.write("tel_001", r0tel)
        writer.write("tel_002", r0tel)  # write a second table too
        writer.write("MC", mc)


def test_read_container(temp_h5_file):
    r0tel1 = R0CameraContainer()
    r0tel2 = R0CameraContainer()
    mc = MCEventContainer()

    reader = HDF5TableReader(str(temp_h5_file))

    # get the generators for each table
    mctab = reader.read('/R0/MC', mc)
    r0tab1 = reader.read('/R0/tel_001', r0tel1)
    r0tab2 = reader.read('/R0/tel_002', r0tel2)

    # read all 3 tables in sync
    for ii in range(3):

        m = next(mctab)
        r0_1 = next(r0tab1)
        r0_2 = next(r0tab2)

        print("MC:", m)
        print("t0:", r0_1.adc_sums)
        print("t1:", r0_2.adc_sums)
        print("---------------------------")

    assert 'test_attribute' in r0_1.meta
    assert r0_1.meta['date'] == "2020-10-10"


def test_read_whole_table(temp_h5_file):

    mc = MCEventContainer()

    reader = HDF5TableReader(str(temp_h5_file))

    for cont in reader.read('/R0/MC', mc):
        print(cont)

if __name__ == '__main__':

    import logging
    logging.basicConfig(level=logging.DEBUG)

    test_write_container("test.h5")
    test_read_container("test.h5")
    test_read_whole_table("test.h5")


"""
Extract data necessary to calcualte charge resolution from raw data files.
"""

import os

import numpy as np
from traitlets import Dict, List, Int, Unicode

from ctapipe.analysis.camera.chargeresolution import ChargeResolutionCalculator
from ctapipe.calib.camera.dl0 import CameraDL0Reducer
from ctapipe.calib.camera.dl1 import CameraDL1Calibrator
from ctapipe.calib.camera.r1 import CameraR1CalibratorFactory
from ctapipe.core import Tool
from ctapipe.image.charge_extractors import ChargeExtractorFactory
from ctapipe.io.eventfilereader import HessioFileReader

from tqdm import tqdm

class ChargeResolutionGenerator(Tool):
    name = "ChargeResolutionGenerator"
    description = "Generate the a pickle file of ChargeResolutionFile for " \
                  "a MC file."

    telescopes = List(Int, None, allow_none=True,
                      help='Telescopes to include from the event file. '
                           'Default = All telescopes').tag(config=True)
    output_name = Unicode('charge_resolution',
                          help='Name of the output charge resolution pickle '
                               'file').tag(config=True)

    aliases = Dict(dict(f='HessioFileReader.input_path',
                        max_events='HessioFileReader.max_events',
                        extractor='ChargeExtractorFactory.extractor',
                        window_width='ChargeExtractorFactory.window_width',
                        window_start='ChargeExtractorFactory.window_start',
                        window_shift='ChargeExtractorFactory.window_shift',
                        sig_amp_cut_HG='ChargeExtractorFactory.sig_amp_cut_HG',
                        sig_amp_cut_LG='ChargeExtractorFactory.sig_amp_cut_LG',
                        lwt='ChargeExtractorFactory.lwt',
                        clip_amplitude='CameraDL1Calibrator.clip_amplitude',
                        radius='CameraDL1Calibrator.radius',
                        max_pe='ChargeResolutionCalculator.max_pe',
                        T='ChargeResolutionGenerator.telescopes',
                        O='ChargeResolutionGenerator.output_name',
                        ))
    classes = List([HessioFileReader,
                    ChargeExtractorFactory,
                    CameraDL1Calibrator,
                    ChargeResolutionCalculator
                    ])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.file_reader = None
        self.r1 = None
        self.dl0 = None
        self.dl1 = None
        self.calculator = None

    def setup(self):
        self.log_format = "%(levelname)s: %(message)s [%(name)s.%(funcName)s]"
        kwargs = dict(config=self.config, tool=self)

        self.file_reader = HessioFileReader(**kwargs)

        extractor_factory = ChargeExtractorFactory(**kwargs)
        extractor_class = extractor_factory.get_class()
        extractor = extractor_class(**kwargs)

        r1_factory = CameraR1CalibratorFactory(origin=self.file_reader.origin,
                                               **kwargs)
        r1_class = r1_factory.get_class()
        self.r1 = r1_class(**kwargs)

        self.dl0 = CameraDL0Reducer(**kwargs)

        self.dl1 = CameraDL1Calibrator(extractor=extractor, **kwargs)

        self.calculator = ChargeResolutionCalculator(**kwargs)

    def start(self):
        desc = "Filling Charge Resolution"
        with tqdm(None, message=desc) as pbar:
            source = self.file_reader.read()
            for event in source:
                pbar.update(event.count)
                tels = list(event.dl0.tels_with_data)

                # Check events have true charge included
                if event.count == 0:
                    try:
                        tel = event.mc.tel[tels[0]]
                        if np.all(tel.photo_electron_image == 0):
                            raise KeyError
                    except KeyError:
                        self.log.exception('Source does not contain '
                                           'true charge!')
                        raise

                self.r1.calibrate(event)
                self.dl0.reduce(event)
                self.dl1.calibrate(event)

                if self.telescopes:
                    tels = []
                    for tel in self.telescopes:
                        if tel in event.dl0.tels_with_data:
                            tels.append(tel)

                for telid in tels:
                    true_charge = event.mc.tel[telid].photo_electron_image
                    measured_charge = event.dl1.tel[telid].image[0]
                    self.calculator.add_charges(true_charge, measured_charge)

    def finish(self):
        directory = self.file_reader.output_directory
        name = "{}.pickle".format(self.output_name)
        ouput_path = os.path.join(directory, name)
        self.calculator.save(ouput_path)


def main():
    exe = ChargeResolutionGenerator()
    exe.run()


if __name__ == '__main__':
    main()



import logging
from collections import namedtuple
import numpy as np
from astropy import units as u
from numba import jit

from ctapipe.instrument import CameraGeometry

logger = logging.getLogger(__name__)

RotBuffer = namedtuple("RotBuffer",
                       "rot_x,rot_y,x_edges,y_edges,new_geom,rot_angle,pix_rotation,"
                       "x_scale")


def unskew_hex_pixel_grid(pix_x, pix_y, cam_angle=0 * u.deg,
                          base_angle=60 * u.deg):
    """transform the pixel coordinates of a hexagonal image into an
    orthogonal image

    Parameters
    ----------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the hexagonal pixel grid
    cam_angle : astropy.Quantity (default: 0 degrees)
        The skewing is performed along the y-axis, therefore, one of the slanted
        base-vectors needs to be parallel to the y-axis.
        Some camera grids are rotated in a way that this is not the case.
        This needs to be corrected.
    base_angle : astropy.Quantity (default: 60 degrees)
        the skewing angle of the hex-grid. should be 60° for regular hexagons

    Returns
    -------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the slanted, orthogonal pixel grid
    """

    tan_angle = np.tan(base_angle)

    # If the camera-rotation angle is non-zero, create a rotation+sheering
    # matrix for the pixel coordinates
    if cam_angle != 0 * u.deg:
        sin_angle = np.sin(cam_angle)
        cos_angle = np.cos(cam_angle)

        # the correction on the pixel position r can be described by a
        # rotation R around one angle and a sheer S along a certain axis:
        #
        #  r'  = S * R * r
        # (x') = (   1    0) * (cos -sin) * (x) = (    cos         -sin    )
        # * (x)
        # (y')   (-1/tan  1)   (sin  cos)   (y)   (sin-cos/tan  sin/tan+cos)
        #   (y)
        # TODO put that in latex...

        rot_mat = np.array(
            [[cos_angle, -sin_angle],
             [sin_angle - cos_angle / tan_angle,
              sin_angle / tan_angle + cos_angle]])

    else:
        # if we don't rotate the camera, only perform the sheer
        rot_mat = np.array([[1, 0], [-1 / tan_angle, 1]])

    rotated = np.dot(rot_mat, [pix_x.value, pix_y.value])
    rot_x = rotated[0] * pix_x.unit
    rot_y = rotated[1] * pix_x.unit
    return rot_x, rot_y


def reskew_hex_pixel_grid(pix_x, pix_y, cam_angle=0 * u.deg,
                          base_angle=60 * u.deg):
    """skews the orthogonal coordinates back to the hexagonal ones

    Parameters
    ----------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the slanted, orthogonal pixel grid
    cam_angle : astropy.Quantity (default: 0 degrees)
        The skewing is performed along the y-axis, therefore, one of the slanted
        base-vectors needs to be parallel to the y-axis.
        Some camera grids are rotated in a way that this is not the case.
        This needs to be corrected.
    base_angle : astropy.Quantity (default: 60 degrees)
        the skewing angle of the hex-grid. should be 60° for regular hexagons

    Returns
    -------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the hexagonal pixel grid

    """

    tan_angle = np.tan(base_angle)

    # If the camera-rotation angle is non-zero, create a rotation+sheering
    # matrix for the pixel coordinates
    if cam_angle != 0 * u.deg:
        sin_angle = np.sin(cam_angle)
        cos_angle = np.cos(cam_angle)

        # to revert the rotation, we need to find matrices S' and R'
        # S' * S = 1 and R' * R = 1
        # so that
        # r = R' * S' * S * R * r = R' * S'*  r'
        #
        # (x) = ( cos sin) * (  1    0) * (x') = (cos+sin/tan  sin) * (x')
        # (y)   (-sin cos)   (1/tan  1)   (y')   (cos/tan-sin  cos)   (y')
        # TODO put that in latex...

        rot_mat = np.array(
            [[cos_angle + sin_angle / tan_angle, sin_angle],
             [cos_angle / tan_angle - sin_angle, cos_angle]])

    else:
        # if we don't rotate the camera, only perform the sheer
        rot_mat = np.array([[1, 0], [1 / tan_angle, 1]])

    rotated = np.dot(rot_mat, [pix_x.value, pix_y.value])
    rot_x = rotated[0] * pix_x.unit
    rot_y = rotated[1] * pix_x.unit
    return rot_x, rot_y


@jit
def reskew_hex_pixel_from_orthogonal_edges(x_edges, y_edges, square_mask):
    """extracts and skews the pixel coordinates from a 2D orthogonal
    histogram (i.e. the bin-edges) and skews them into the hexagonal
    image while selecting only the pixel that are selected by the
    given mask

    Parameters
    ----------
    x_edges, y_edges : 1darrays
        the bin edges of the 2D histogram
    square_mask : 2darray
        mask that selects the pixels actually belonging to the camera

    Returns
    -------
    unrot_x, unrot_y : 1darrays
        pixel coordinated reskewed into the hexagonal camera grid
    """

    unrot_x, unrot_y = [], []
    for i, x in enumerate((x_edges[:-1] + x_edges[1:]) / 2):
        for j, y in enumerate((y_edges[:-1] + y_edges[1:]) / 2):
            if square_mask[i][j]:
                x_unrot, y_unrot = reskew_hex_pixel_grid(x, y)
                unrot_x.append(x_unrot)
                unrot_y.append(y_unrot)
    return unrot_x, unrot_y


@jit
def get_orthogonal_grid_edges(pix_x, pix_y, scale_aspect=True):
    """calculate the bin edges of the slanted, orthogonal pixel grid to
    resample the pixel signals with np.histogramdd right after.

    Parameters
    ----------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the slanted, orthogonal pixel grid
    scale_aspect : boolean (default: True)
        if True, rescales the x-coordinates to create square pixels
        (instead of rectangular ones)

    Returns
    --------
    x_edges, y_edges : 1D numpy arrays
        the bin edges for the slanted, orthogonal pixel grid
    x_scale : float
        factor by which the x-coordinates have been scaled
    """

    # finding the size of the square patches
    d_x = 99 * u.m
    d_y = 99 * u.m
    x_base = pix_x[0]
    y_base = pix_y[0]
    for x, y in zip(pix_x, pix_y):
        if abs(y - y_base) < abs(x - x_base):
            d_x = min(d_x, abs(x - x_base))
    for x, y in zip(pix_x, pix_y):
        if abs(y - y_base) > abs(x - x_base):
            d_y = min(d_y, abs(y - y_base))

    x_scale = 1
    if scale_aspect:
        x_scale = d_y / d_x
        pix_x *= x_scale
        d_x = d_y

    # with the maximal extension of the axes and the size of the pixels,
    # determine the number of bins in each direction
    NBinsx = np.around(abs(max(pix_x) - min(pix_x)) / d_x) + 2
    NBinsy = np.around(abs(max(pix_y) - min(pix_y)) / d_y) + 2
    x_edges = np.linspace(min(pix_x).value, max(pix_x).value, NBinsx)
    y_edges = np.linspace(min(pix_y).value, max(pix_y).value, NBinsy)

    return x_edges, y_edges, x_scale


add_angle = 180 * u.deg
rot_buffer = {}


def convert_geometry_1d_to_2d(geom, signal, key=None, add_rot=0):
    """converts the geometry object of a camera with a hexagonal grid into
    a square grid by slanting and stretching the 1D arrays of pixel x
    and y positions and signal intensities are converted to 2D
    arrays. If the signal array contains a time-dimension it is
    conserved.

    Parameters
    ----------
    geom : CameraGeometry object
        geometry object of hexagonal cameras
    signal : ndarray
        1D (no timing) or 2D (with timing) array of the pmt signals
    key : (default: None)
        arbitrary key to store the transformed geometry in a buffer
    add_rot : int/float (default: 0)
        parameter to apply an additional rotation of @add_rot times 60°

    Returns
    -------
    new_geom : CameraGeometry object
        geometry object of the slanted picture now with a rectangular
        grid and a 2D grid for the pixel positions contains now a 2D
        masking array signifying which of the pixels came from the
        original geometry and which are simply fillers from the
        rectangular grid square_img : ndarray 2D (no timing) or 3D
        (with timing) array of the pmt signals

    """

    if key in rot_buffer:

        # if the conversion with this key was done and stored before,
        # just read it in
        (rot_x, rot_y, x_edges, y_edges, new_geom,
         rot_angle, pix_rotation, x_scale) = rot_buffer[key]
    else:

        # otherwise, we have to do the conversion now first, skew all
        # the coordinates of the original geometry

        # extra_rot is the angle to get back to aligned hexagons with flat
        # tops. Note that the pixel rotation angle brings the camera so that
        # hexagons have a point at the top, so need to go 30deg back to
        # make them flat
        extra_rot = geom.pix_rotation - 30*u.deg

        # total rotation angle:
        rot_angle = (add_rot * 60 * u.deg) - extra_rot
        # if geom.cam_id.startswith("NectarCam")\
        #         or geom.cam_id.startswith("LSTCam"):
        #     rot_angle += geom.cam_rotation + 90 * u.deg

        logger.debug("geom={}".format(geom))
        logger.debug("rot={}, extra={}".format(rot_angle, extra_rot))

        rot_x, rot_y = unskew_hex_pixel_grid(geom.pix_x, geom.pix_y,
                                             cam_angle=rot_angle)

        # with all the coordinate points, we can define the bin edges
        # of a 2D histogram
        x_edges, y_edges, x_scale = get_orthogonal_grid_edges(rot_x, rot_y)

        # this histogram will introduce bins that do not correspond to
        # any pixel from the original geometry. so we create a mask to
        # remember the true camera pixels by simply throwing all pixel
        # positions into numpy.histogramdd: proper pixels contain the
        # value 1, false pixels the value 0.
        square_mask = np.histogramdd([rot_y, rot_x],
                                     bins=(y_edges, x_edges))[0].astype(bool)

        # to be consistent with the pixel intensity, instead of saving
        # only the rotated positions of the true pixels (rot_x and
        # rot_y), create 2D arrays of all x and y positions (also the
        # false ones).
        grid_x, grid_y = np.meshgrid((x_edges[:-1] + x_edges[1:]) / 2.,
                                     (y_edges[:-1] + y_edges[1:]) / 2.)

        ids = []
        # instead of blindly enumerating all pixels, let's instead
        # store a list of all valid -- i.e. picked by the mask -- 2D
        # indices
        for i, row in enumerate(square_mask):
            for j, val in enumerate(row):
                if val is True:
                    ids.append((i, j))

        # the area of the pixels (note that this is still a deformed
        # image)
        pix_area = np.ones_like(grid_x) \
                   * (x_edges[1] - x_edges[0]) * (y_edges[1] - y_edges[0])

        # creating a new geometry object with the attributes we just determined
        new_geom = CameraGeometry(
            cam_id=geom.cam_id+"_rect",
            pix_id=ids,  # this is a list of all the valid coordinate pairs now
            pix_x=grid_x * u.m,
            pix_y=grid_y * u.m,
            pix_area=pix_area * u.m ** 2,
            neighbors=geom.neighbors,
            pix_type='rectangular', apply_derotation=False)

        # storing the pixel mask and camera rotation for later use
        new_geom.mask = square_mask

        if key is not None:
            # if a key is given, store the essential objects in a buffer
            rot_buffer[key] = RotBuffer(rot_x, rot_y, x_edges, y_edges, new_geom,
                                        rot_angle, geom.pix_rotation, x_scale)

    # resample the signal array to correspond to the square grid --
    #  for signal arrays containing time slices (ndim > 1) or not
    #  approach is the same as used for the mask only with the signal
    #  as bin-weights
    if signal.ndim > 1:
        t_dim = signal.shape[1]
        square_img = np.histogramdd([np.repeat(rot_y, t_dim),
                                     np.repeat(rot_x, t_dim),
                                     [a for a in range(t_dim)] * len(rot_x)],
                                    bins=(y_edges, x_edges, range(t_dim + 1)),
                                    weights=signal.ravel())[0]
    else:
        square_img = np.histogramdd([rot_y, rot_x],
                                    bins=(y_edges, x_edges),
                                    weights=signal)[0]

    return new_geom, square_img


unrot_buffer = {}  # todo: should just make this a function with @lru_cache


def convert_geometry_back(geom, signal, key, add_rot=0):
    """reverts the geometry distortion performed by
    convert_geometry_1d_to_2d back to a hexagonal grid stored in 1D
    arrays

    Parameters
    ----------
    geom : CameraGeometry
        geometry object where pixel positions are stored in a 2D
        rectangular camera grid
    signal : ndarray
        pixel intensity stored in a 2D rectangular camera grid
    key:
        key to retrieve buffered geometry information
    add_rot : int/float (default: 0)
        parameter to apply an additional rotation of @add_rot times 60°

    Returns
    -------
    unrot_geom : CameraGeometry
        pixel rotated back to a hexagonal grid stored in a 1D array
    signal : ndarray
        1D (no timing) or 2D (with timing) array of the pmt signals

    """
    global unrot_buffer

    square_mask = geom.mask

    if key in unrot_buffer:
        unrot_geom = unrot_buffer[key]
    else:
        if key in rot_buffer:
            x_scale = rot_buffer[key].x_scale
            rot_angle = rot_buffer[key].rot_angle
            pix_rotation = rot_buffer[key].pix_rotation
        else:
            raise KeyError("key '{}' not found in the buffer".format(key))

        grid_x, grid_y = geom.pix_x / x_scale, geom.pix_y

        unrot_x, unrot_y = reskew_hex_pixel_grid(grid_x[square_mask],
                                                 grid_y[square_mask],
                                                 rot_angle)

        # TODO: probably should use base constructor, not guess here:
        # unrot_geom = CameraGeometry.guess(unrot_x, unrot_y, foc_len,
        #                                  apply_derotation=False)
        unrot_geom = CameraGeometry(cam_id=geom.cam_id + "_hex",
                                    pix_id=np.arange(len(unrot_x)),
                                    pix_x=unrot_x,
                                    pix_y=unrot_y,
                                    pix_area=None,  # recalc
                                    pix_type='hexagonal',
                                    pix_rotation=pix_rotation,
                                    # apply_derotation=False
                                    )

        unrot_buffer[key] = unrot_geom

    return unrot_geom, signal[square_mask, ...]




"""
from ctapipe.flow.gui.main_window import ModuleApplication
from ctapipe.flow.gui.graphwidget import GraphWidget
from ctapipe.flow.gui.infolabel import *
from ctapipe.flow.gui.guiconnection  import *


all = ['GraphWidget', 'Ui_MainWindow', 'ModuleApplication',
       'GuiConnexion', 'StagerRep', 'ProducerRep', 'ConsumerRep','InfoLabel']
"""



"""This module defines the important coordinate systems to be used in
reconstruction with the CTA pipeline and the transformations between
this different systems. Frames and transformations are defined using
the astropy.coordinates framework. This module defines transformations
for ground based cartesian and planar systems.

For examples on usage see examples/coordinate_transformations.py

This code is based on the coordinate transformations performed in the
read_hess code

TODO:

- Tests Tests Tests!
"""

import astropy.units as u
import numpy as np
from astropy.coordinates import (BaseCoordinateFrame,
                                 CartesianRepresentation,
                                 FunctionTransform)

try:
    # FrameAttribute was renamed Attribute in astropy 2.0
    # TODO: should really use subclasses like QuantityAttribute
    from astropy.coordinates import FrameAttribute as Attribute
except ImportError:
    from astropy.coordinates import Attribute


from astropy.coordinates import frame_transform_graph
from numpy import cos, sin
from ..coordinates.representation import PlanarRepresentation

__all__ = [
    'GroundFrame',
    'TiltedGroundFrame',
    'project_to_ground'
]


class GroundFrame(BaseCoordinateFrame):
    """Ground coordinate frame.  The ground coordinate frame is a simple
    cartesian frame describing the 3 dimensional position of objects
    compared to the array ground level in relation to the nomial
    centre of the array.  Typically this frame will be used for
    describing the position on telescopes and equipment

    Frame attributes: None

    """
    default_representation = CartesianRepresentation
    # Pointing direction of the tilted system (alt,az),
    # could be the telescope pointing direction or the reconstructed shower
    # direction
    pointing_direction = Attribute(default=None)


class TiltedGroundFrame(BaseCoordinateFrame):
    """Tilted ground coordinate frame.  The tilted ground coordinate frame
    is a cartesian system describing the 2 dimensional projected
    positions of objects in a tilted plane described by
    pointing_direction Typically this frame will be used for the
    reconstruction of the shower core position

    Frame attributes:
    
    * ``pointing_direction``
        Alt,Az direction of the tilted reference plane

    """
    default_representation = PlanarRepresentation
    # Pointing direction of the tilted system (alt,az),
    # could be the telescope pointing direction or the reconstructed shower
    # direction
    pointing_direction = Attribute(default=None)

# Transformations defined below this point


def get_shower_trans_matrix(azimuth, altitude):
    """Get Transformation matrix for conversion from the ground system to
    the Tilted system and back again (This function is directly lifted
    from read_hess, probably could be streamlined using python
    functionality)

    Parameters
    ----------
    azimuth: float
        Azimuth angle of the tilted system used
    altitude: float
        Altitude angle of the tilted system used

    Returns
    -------
    trans: 3x3 ndarray transformation matrix
    """

    cos_z = sin(altitude)
    sin_z = cos(altitude)
    cos_az = cos(azimuth)
    sin_az = sin(azimuth)

    trans = np.zeros([3, 3])
    trans[0][0] = cos_z * cos_az
    trans[1][0] = sin_az
    trans[2][0] = sin_z * cos_az

    trans[0][1] = -cos_z * sin_az
    trans[1][1] = cos_az
    trans[2][1] = -sin_z * sin_az

    trans[0][2] = -sin_z
    trans[1][2] = 0.
    trans[2][2] = cos_z

    return trans


@frame_transform_graph.transform(FunctionTransform, GroundFrame,
                                 TiltedGroundFrame)
def ground_to_tilted(ground_coord, tilted_coord):
    """
    Transformation from ground system to tilted ground system

    Parameters
    ----------
    ground_coord: `astropy.coordinates.SkyCoord`
        GroundFrame system
    tilted_coord: `astropy.coordinates.SkyCoord`
        TiltedGroundFrame system

    Returns
    -------
    TiltedGroundFrame coordinates
    """
    x_grd = ground_coord.cartesian.x
    y_grd = ground_coord.cartesian.y
    z_grd = ground_coord.cartesian.z

    altitude, azimuth = tilted_coord.pointing_direction.alt, tilted_coord.pointing_direction.az
    altitude = altitude.to(u.rad)
    azimuth = azimuth.to(u.rad)
    trans = get_shower_trans_matrix(azimuth, altitude)

    x_tilt = trans[0][0] * x_grd + trans[0][1] * y_grd + trans[0][2] * z_grd
    y_tilt = trans[1][0] * x_grd + trans[1][1] * y_grd + trans[1][2] * z_grd

    representation = PlanarRepresentation(x_tilt, y_tilt)

    return tilted_coord.realize_frame(representation)


@frame_transform_graph.transform(FunctionTransform, TiltedGroundFrame,
                                 GroundFrame)
def tilted_to_ground(tilted_coord, ground_coord):
    """
    Transformation from tilted ground system to  ground system

    Parameters
    ----------
    tilted_coord: `astropy.coordinates.SkyCoord`
        TiltedGroundFrame system
    ground_coord: `astropy.coordinates.SkyCoord`
        GroundFrame system

    Returns
    -------
    GroundFrame coordinates
    """
    x_tilt = tilted_coord.x
    y_tilt = tilted_coord.y

    altitude, azimuth = tilted_coord.pointing_direction.alt, tilted_coord.pointing_direction.az
    altitude = altitude.to(u.rad)
    azimuth = azimuth.to(u.rad)

    trans = get_shower_trans_matrix(azimuth, altitude)

    x_grd = trans[0][0] * x_tilt + trans[1][0] * y_tilt
    y_grd = trans[0][1] * x_tilt + trans[1][1] * y_tilt
    z_grd = trans[0][2] * x_tilt + trans[1][2] * y_tilt

    representation = CartesianRepresentation(x_grd, y_grd, z_grd)

    grd = ground_coord.realize_frame(representation)
    return grd


def project_to_ground(tilt_system):
    """Project position in the tilted system onto the ground. This is
    needed as the standard transformation will return the 3d position
    of the tilted frame. This projection may ultimately be the
    standard use case so may be implemented in the tilted to ground
    transformation

    Parameters
    ----------
    tilt_system: `astropy.coordinates.SkyCoord`
        coorinate in the the tilted ground system

    Returns
    -------
    Projection of tilted system onto the ground (GroundSystem)

    """
    ground_system = tilt_system.transform_to(GroundFrame)

    unit = ground_system.x.unit
    x_initial = ground_system.x.value
    y_initial = ground_system.y.value
    z_initial = ground_system.z.value

    trans = get_shower_trans_matrix(tilt_system.pointing_direction.az,
                                    tilt_system.pointing_direction.alt)

    x_projected = x_initial - trans[2][0] * z_initial / trans[2][2]
    y_projected = y_initial - trans[2][1] * z_initial / trans[2][2]

    return GroundFrame(x=x_projected * unit, y=y_projected * unit, z=0 * unit)



""""
Line-intersection-based fitting.

Contact: Tino Michael <Tino.Michael@cea.fr>
"""

from ctapipe.utils import linalg
from ctapipe.reco.reco_algorithms import Reconstructor
from ctapipe.io.containers import ReconstructedShowerContainer

from astropy.utils.decorators import deprecated

from itertools import combinations

import numpy as np

from scipy.optimize import minimize

from astropy import units as u
u.dimless = u.dimensionless_unscaled


__all__ = ['HillasReconstructor',
           'TooFewTelescopesException',
           'dist_to_traces', 'MEst', 'GreatCircle']


class TooFewTelescopesException(Exception):
    pass


@deprecated(0.1, "will be replaced with real coord transform")
def guess_pix_direction(pix_x, pix_y, tel_phi, tel_theta, tel_foclen):
    """
    TODO replace with proper implementation
    calculates the direction vector of corresponding to a
    (x,y) position on the camera

    beta is the pixel's angular distance to the centre
    according to beta / tel_view = r / maxR
    alpha is the polar angle between the y-axis and the pixel
    to find the direction the pixel is looking at:

    - the pixel direction is set to the telescope direction
    - offset by beta towards up
    - rotated around the telescope direction by the angle alpha


    Parameters
    -----------
    pix_x, pix_y : ndarray
        lists of x and y positions on the camera
    tel_phi, tel_theta: astropy quantities
        two angles that describe the orientation of the telescope
    tel_foclen : astropy quantity
        focal length of the telescope

    Returns
    -------
    pix_dirs : ndarray
        shape (n,3) list of "direction vectors"
        corresponding to a position on the camera

    """

    pix_alpha = np.arctan2(pix_y, pix_x)
    pix_rho = (pix_x ** 2 + pix_y ** 2) ** .5

    pix_beta = pix_rho / tel_foclen * u.rad

    tel_dir = linalg.set_phi_theta(tel_phi, tel_theta)

    pix_dirs = []
    for a, b in zip(pix_alpha, pix_beta):
        pix_dir = linalg.set_phi_theta(tel_phi, tel_theta + b)

        pix_dir = linalg.rotate_around_axis(pix_dir, tel_dir, -a)
        pix_dirs.append(pix_dir * u.dimless)

    return pix_dirs


def dist_to_traces(core, circles):
    """This function calculates the M-Estimator from the distances of the
    suggested core position to all traces of the given GreatCircles.
    The distance of the core to the trace line is the length of the
    vector between the core and an arbitrary point on the trace
    projected perpendicular to the trace.

    This is implemented as the scalar product of the connecting vector
    between the core and the position of the telescope and { trace[1],
    -trace[0] } as the normal vector of the trace.

    Notes
    -----
    uses the M-Estimator of the distance instead of the distance itself:

    .. math::

        M_\\text{Est} = \sum_i{ 2 \cdot \sqrt{1 + d_i^2} - 2}


    """

    mest = 0.
    for circle in circles.values():

        # the distanece of the core
        D = core - circle.pos[:2] / u.m
        dist = D[0] * circle.trace[1] - D[1] * circle.trace[0]

        # summing up the M-Estimator with the given circle weights
        mest += (2 * np.sqrt(1 + (dist ** 2)) - 2) * circle.weight
    return mest


def MEst(origin, circles, weights):
    """calculates the M-Estimator: a modified χ² that becomes
    asymptotically linear for high values and is therefore less
    sensitive to outliers

    the test is performed to maximise the angles between the fit
    direction and the all the normal vectors of the great circles

    .. math::

        M_\\text{Est} = \sum_i{ 2 \cdot \sqrt{1 + d_i^2} - 2}


    Notes
    -----
    seemingly inferior to negative sum of sin(angle)...


    Parameters
    -----------
    origin : length-3 array
        direction vector of the gamma's origin used as seed
    circles : GreatCircle array
        collection of great circles created from the camera images
    weights : array
        list of weights for each image/great circle

    Returns
    -------
    MEstimator : float


    """

    sin_ang = np.array([linalg.length(np.cross(origin, circ.norm))
                        for circ in circles.values()])
    return -2 * np.sum(weights * np.sqrt((1 + np.square(sin_ang))) - 2)


def neg_angle_sum(origin, circles, weights):
    """calculates the negative sum of the angle between the fit direction
    and all the normal vectors of the great circles

    Parameters
    -----------
    origin : length-3 array
        direction vector of the gamma's origin used as seed
    circles : GreatCircle array
        collection of great circles created from the camera images
    weights : array
        list of weights for each image/great circle

    Returns
    --------
    n_sum_angles : float
        negative of the sum of the angles between the test direction
        and all normal vectors of the given great circles

    """

    sin_ang = np.array([linalg.length(np.cross(origin, circ.norm))
                        for circ in circles.values()])
    return -np.sum(weights * sin_ang)


class HillasReconstructor(Reconstructor):
    """
    class that reconstructs the direction of an atmospheric shower
    using a simple hillas parametrisation of the camera images it
    provides a direction estimate in two steps and an estimate for the
    shower's impact position on the ground.

    so far, it does neither provide an energy estimator nor an
    uncertainty on the reconstructed parameters

    """

    def __init__(self, configurable=None):
        super().__init__(configurable)
        self.circles = {}

    def predict(self, hillas_dict, inst, tel_phi, tel_theta, seed_pos=(0, 0)):
        """The function you want to call for the reconstruction of the
        event. It takes care of setting up the event and consecutively
        calls the functions for the direction and core position
        reconstruction.  Shower parameters not reconstructed by this
        class are set to np.nan

        Parameters
        -----------
        hillas_dict : python dictionary
            dictionary with telescope IDs as key and
            MomentParameters instances as values
        seed_pos : python tuple
            shape (2) tuple with a possible seed for
            the core position fit (e.g. CoG of all telescope images)

        Raises
        ------
        TooFewTelescopesException
            if len(hillas_dict) < 2

        """

        # stereoscopy needs at least two telescopes
        if len(hillas_dict) < 2:
            raise TooFewTelescopesException(
                "need at least two telescopes, have {}"
                .format(len(hillas_dict)))

        self.get_great_circles(hillas_dict, inst.subarray, tel_phi, tel_theta)

        # algebraic direction estimate
        dir = self.fit_origin_crosses()[0]

        # core position estimate using a geometric approach
        pos, pos_uncert = self.fit_core_crosses()

        # numerical minimisations do not really improve the fit
        # direction estimate using numerical minimisation
        # dir = self.fit_origin_minimise(dir)
        #
        # core position estimate using numerical minimisation
        # pos = self.fit_core_minimise(seed_pos)

        # container class for reconstructed showers
        result = ReconstructedShowerContainer()
        phi, theta = linalg.get_phi_theta(dir).to(u.deg)

        # TODO make sure az and phi turn in same direction...
        result.alt, result.az = 90 * u.deg - theta, phi
        result.core_x = pos[0]
        result.core_y = pos[1]
        result.core_uncert = pos_uncert

        result.tel_ids = [h for h in hillas_dict.keys()]
        result.average_size = np.mean([h.size for h in hillas_dict.values()])
        result.is_valid = True

        result.alt_uncert = np.nan
        result.az_uncert = np.nan
        result.h_max = np.nan
        result.h_max_uncert = np.nan
        result.goodness_of_fit = np.nan

        return result

    def get_great_circles(self, hillas_dict, subarray, tel_phi, tel_theta):
        """
        creates a dictionary of :class:`.GreatCircle` from a dictionary of
        hillas
        parameters

        Parameters
        ----------
        hillas_dict : dictionary
            dictionary of hillas moments
        subarray : ctapipe.instrument.SubarrayDescription
            subarray information
        tel_phi, tel_theta : dictionaries
            dictionaries of the orientation angles of the telescopes
            needs to contain at least the same keys as in `hillas_dict`
        """

        self.circles = {}
        for tel_id, moments in hillas_dict.items():

            p2_x = moments.cen_x + moments.length * np.cos(moments.psi)
            p2_y = moments.cen_y + moments.length * np.sin(moments.psi)
            foclen = subarray.tel[tel_id].optics.effective_focal_length

            circle = GreatCircle(
                guess_pix_direction(np.array([moments.cen_x / u.m,
                                              p2_x / u.m]) * u.m,
                                    np.array([moments.cen_y / u.m,
                                              p2_y / u.m]) * u.m,
                                    tel_phi[tel_id], tel_theta[tel_id],
                                    foclen),
                moments.size * (moments.length / moments.width)
            )
            circle.pos = subarray.positions[tel_id]
            self.circles[tel_id] = circle

    def fit_origin_crosses(self):
        """calculates the origin of the gamma as the weighted average
        direction of the intersections of all great circles

        Returns
        -------
        gamma : shape (3) numpy array
            direction of origin of the reconstructed shower as a 3D vector
        crossings : shape (n,3) list
            list of all the crossings of the `GreatCircle` list
        """

        crossings = []
        for perm in combinations(self.circles.values(), 2):
            n1, n2 = perm[0].norm, perm[1].norm
            # cross product automatically weighs in the angle between
            # the two vectors: narrower angles have less impact,
            # perpendicular vectors have the most
            crossing = np.cross(n1, n2)

            # two great circles cross each other twice (one would be
            # the origin, the other one the direction of the gamma) it
            # doesn't matter which we pick but it should at least be
            # consistent: make sure to always take the "upper" solution
            if crossing[2] < 0:
                crossing *= -1
            crossings.append(crossing * perm[0].weight * perm[1].weight)

        # averaging over the solutions of all permutations
        return linalg.normalise(np.sum(crossings, axis=0)) * u.dimless, crossings

    def fit_origin_minimise(self, seed=(0, 0, 1), test_function=neg_angle_sum):
        """ Fits the origin of the gamma with a minimisation procedure this
        function expects that `get_great_circles`
        has been run already. A seed should be given otherwise it defaults to
        "straight up" supperted functions to minimise are an M-estimator and the
        negative sum of the angles to all normal vectors of the great
        circles

        Parameters
        -----------
        seed : length-3 array
            starting point of the minimisation
        test_function : function object, optional (default: neg_angle_sum)
            either neg_angle_sum or MEst (or own implementation...)
            neg_angle_sum seemingly superior to MEst

        Returns
        -------
        direction : length-3 numpy array as dimensionless quantity
            best fit for the origin of the gamma from
            the minimisation process

        """

        # using the sum of the cosines of each direction with every
        # other direction as weight; don't use the product -- with many
        # steep angles, the product will become too small and the
        # weight (and the whole fit) useless
        weights = [np.sum([linalg.length(np.cross(A.norm, B.norm))
                           for A in self.circles.values()]
                          ) * B.weight
                   for B in self.circles.values()]

        # minimising the test function
        self.fit_result_origin = minimize(test_function, seed,
                                          args=(self.circles, weights),
                                          method='BFGS',
                                          options={'disp': False}
                                          )

        return np.array(linalg.normalise(self.fit_result_origin.x)) * u.dimless

    def fit_core_crosses(self, unit=u.m):
        r"""calculates the core position as the least linear square solution
        of an (over-constrained) equation system

        Notes
        -----
        The basis is the "trace" of each telescope's `GreatCircle` which
        can be determined by the telescope's position P=(Px, Py) and
        the circle's normal vector, projected to the ground n=(nx,
        ny), so that for every r=(x, y) on the trace

        :math:`\vec n \cdot \vec r = \vec n \cdot \vec P` ,

        :math:`n_x \cdot x + n_y \cdot y = d`

        In a perfect world, the traces of all telescopes cross in the
        shower's point of impact. This means that there is one common
        point (x, y) for every telescope, so we can write in matrix
        form:

        .. math::
            :label: fullmatrix

            \begin{pmatrix}
                nx_1  &  ny_1  \\
                \vdots & \vdots \\
                nx_n  &  ny_n
            \end{pmatrix}
                \cdot (x, y) =
            \begin{pmatrix}
                d_1  \\
                \vdots \\
                d_n
            \end{pmatrix}



        or :math:`\boldsymbol{A} \cdot \vec r = \vec D` .

        Since we do not live in a perfect world and there probably is
        no point r that fulfils this equation system, it is solved by
        the method of least linear square:

        .. math::
            :label: rchisqr

            \vec{r}_{\chi^2} = (\boldsymbol{A}^\text{T} \cdot \boldsymbol{A})^{-1}
            \boldsymbol{A}^\text{T} \cdot \vec D


        :math:`\vec{r}_{\chi^2}` minimises the squared difference of


        .. math::

            \vec D - \boldsymbol{A} \cdot \vec r


        Weights are applied to every line of equation :eq:`fullmatrix`
        as stored in circle.weight (assuming they have been set in
        `get_great_circles` or elsewhere).

        Returns
        -------
        r_chisqr: numpy.ndarray(2)
            the minimum :math:`\chi^2` solution for the shower impact position
        pos_uncert: astropy length quantity
            error estimate on the reconstructed core position

        """

        A = np.zeros((len(self.circles), 2))
        D = np.zeros(len(self.circles))
        for i, circle in enumerate(self.circles.values()):
            # apply weight from circle and from the tilt of the circle
            # towards the horizontal plane: simply projecting
            # circle.norm to the ground gives higher weight to planes
            # perpendicular to the ground and less to those that have
            # a steeper angle
            A[i] = circle.weight * circle.norm[:2]
            # since A[i] is used in the dot-product, no need to multiply the
            # weight here
            D[i] = np.dot(A[i], circle.pos[:2])

        # the math from equation (2) would look like this:
        # ATA = np.dot(A.T, A)
        # ATAinv = np.linalg.inv(ATA)
        # ATAinvAT = np.dot(ATAinv, A.T)
        # return np.dot(ATAinvAT, D) * unit

        # instead used directly the numpy implementation
        # speed is the same, just handles already "SingularMatrixError"
        pos = np.linalg.lstsq(A, D)[0] * u.m

        weighted_sum_dist = np.sum([np.dot(pos[:2] - c.pos[:2], c.norm[:2]) * c.weight
                                    for c in self.circles.values()]) * pos.unit
        norm_sum_dist = np.sum([c.weight * linalg.length(c.norm[:2])
                                for c in self.circles.values()])
        pos_uncert = abs(weighted_sum_dist / norm_sum_dist)

        return pos, pos_uncert

    def fit_core_minimise(self, seed=(0, 0), test_function=dist_to_traces):
        """
        reconstructs the shower core position from the already set up
        great circles

        Notes
        -----
        The core of the shower lies on the cross section of the great
        circle with the horizontal plane. The direction of this cross
        section is the cross-product of the circle's normal vector and
        the horizontal plane.  Here, we only care about the direction;
        not the orientation...


        Parameters
        ----------
        seed : tuple
            shape (2) tuple with optional starting coordinates
            tuple of floats or astropy.length -- if floats, assume metres
        test_function : function object, optional (default: dist_to_traces)
            function to be used by the minimiser

        """

        if type(seed) == u.Quantity:
            unit = seed.unit
        else:
            unit = u.m

        # the direction of the cross section of the great circle with
        # the horizontal frame is the cross product of the great
        # circle's normal vector with the z-axis:
        # n × z = (n[1], -n[0], 0)
        for circle in self.circles.values():
            circle.trace = linalg.normalise(np.array([circle.norm[1],
                                                      -circle.norm[0], 0]))

        # minimising the test function (note: minimize strips seed off its
        # unit)
        self.fit_result_core = minimize(test_function, seed[:2],
                                        args=(self.circles),
                                        method='BFGS', options={'disp': False})

        if not self.fit_result_core.success:
            print("fit_core: fit no success")

        return np.array(self.fit_result_core.x) * unit


class GreatCircle:
    """
    a tiny helper class to collect some parameters for each great great
    circle
    """

    def __init__(self, dirs, weight=1):
        """the constructor takes two directions on the circle and creates the
        normal vector belonging to that plane

        Parameters
        -----------
        dirs : shape (2,3) ndarray
            contains two 3D direction-vectors
        weight : float, optional
            weight of this telescope for later use during the
            reconstruction

        Notes
        -----
        c: numpy.ndarray(3)
            :math:`\vec c = (\vec a \times \vec b) \times \vec a`
            :math:`\rightarrow` a and c form an orthogonal base for the 
            great circle
            (only orthonormal if a and b are of unit-length)
        norm: numpy.ndarray(3)
            normal vector of the circle's plane,
            perpendicular to a, b and c
        """

        self.a = dirs[0]
        self.b = dirs[1]

        # a and c form an orthogonal basis for the great circle
        # not really necessary since the norm can be calculated
        # with a and b just as well
        self.c = np.cross(np.cross(self.a, self.b), self.a)
        # normal vector for the plane defined by the great circle
        self.norm = linalg.normalise(np.cross(self.a, self.c))
        # some weight for this circle
        # (put e.g. uncertainty on the Hillas parameters
        # or number of PE in here)
        self.weight = weight



"""
Functions to retrieve and interpolate atmosphere profiles.
"""

import numpy as np
from astropy.table import Table
from astropy.units import Quantity
from scipy.interpolate import interp1d
from ctapipe.core.provenance import Provenance

from ctapipe.utils import get_dataset

__all__ = ['get_atmosphere_profile_table', 'get_atmosphere_profile_functions']


def get_atmosphere_profile_table(atmosphere_name='paranal'):
    """
    Get an atmosphere profile table
    
    Parameters
    ----------
    atmosphere_name: str
        identifier of atmosphere profile

    Returns
    -------
    astropy.table.Table  containing atmosphere profile with at least columns 
    'altitude' (m), and 'thickness' (g cm-2) as well as others.

    """
    filename = '{}.atmprof.fits.gz'.format(atmosphere_name)
    table = Table.read(get_dataset(filename))
    Provenance().add_input_file(filename, role='dl0.arr.svc.atmosphere')
    return table


def get_atmosphere_profile_functions(atmosphere_name="paranal",
                                     with_units=True):
    """ 
    Gives atmospheric profile as a continuous function thickness(
    altitude), and it's inverse altitude(thickness)  in m and g/cm^2 
    respectively.
    
    Parameters
    ----------
    atmosphere_name: str
        identifier of atmosphere profile
    with_units: bool
       if true, return functions that accept and return unit quantities. 
       Otherwise assume units are 'm' and 'g cm-2'

    Returns
    -------
    functions: thickness(alt), alt(thickness)
    """
    tab = get_atmosphere_profile_table(atmosphere_name)
    alt = tab['altitude'].to('m')
    thick = (tab['thickness']).to("g cm-2")

    alt_to_thickness = interp1d(x=np.array(alt), y=np.array(thick))
    thickness_to_alt = interp1d(x=np.array(thick), y=np.array(alt))

    if with_units:
        def thickness(a):
            return Quantity(alt_to_thickness(a.to('m')), 'g cm-2')

        def altitude(a):
            return Quantity(thickness_to_alt(a.to('g cm-2')), 'm')

        return thickness, altitude

    return  alt_to_thickness, thickness_to_alt


"""
Handles reading of different event/waveform containing files
"""
from abc import abstractmethod
from os.path import basename, splitext, dirname, join, exists
import numpy as np
from traitlets import Unicode, Int, CaselessStrEnum, observe
from copy import deepcopy
from ctapipe.core import Component, Factory
from ctapipe.utils import get_dataset
from ctapipe.io.hessio import hessio_event_source, hessio_get_list_event_ids


class EventFileReader(Component):
    """
    Parent class for specific FileReaders

    Attributes
    ----------
    input_path : str
    directory : str
        Automatically set from `input_path`.
    filename : str
        Name of the file without the extension.
        Automatically set from `input_path`.
    extension : str
        Automatically set from `input_path`.
    output_directory : str
        Directory to save outputs for this file

    """
    name = 'EventFileReader'
    origin = None

    input_path = Unicode(get_dataset('gamma_test.simtel.gz'), allow_none=True,
                         help='Path to the input file containing '
                              'events.').tag(config=True)
    max_events = Int(None, allow_none=True,
                     help='Maximum number of events that will be read from'
                          'the file').tag(config=True)

    def __init__(self, config, tool, **kwargs):
        """
        Class to handle generic input files. Enables obtaining the "source"
        generator, regardless of the type of file (either hessio or camera
        file).

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)

        if self.origin is None:
            raise ValueError("Subclass of EventFileReader should specify "
                             "an origin")

        self._num_events = None
        self._event_id_list = []

        if self.input_path is None:
            raise ValueError("Please specify an input_path for event file")
        self._init_path(self.input_path)

    def _init_path(self, input_path):
        if not exists(input_path):
            raise FileNotFoundError("file path does not exist: '{}'"
                                    .format(input_path))

        self.input_path = input_path
        self.directory = dirname(input_path)
        self.filename = splitext(basename(input_path))[0]
        self.extension = splitext(input_path)[1]
        self.output_directory = join(self.directory, self.filename)

        if self.log:
            self.log.info("INPUT PATH = {}".format(self.input_path))

    @observe('input_path')
    def on_input_path_changed(self, change):
        new = change['new']
        try:
            self.log.warning("Change: input_path={}".format(new))
            self._num_events = None
            self._event_id_list = []
            self._init_path(new)
        except AttributeError:
            pass

    @observe('origin')
    def on_origin_changed(self, change):
        new = change['new']
        try:
            self.log.warning("Change: origin={}".format(new))
        except AttributeError:
            pass

    @observe('max_events')
    def on_max_events_changed(self, change):
        new = change['new']
        try:
            self.log.warning("Change: max_events={}".format(new))
            self._num_events = None
            self._event_id_list = []
        except AttributeError:
            pass

    @property
    @abstractmethod
    def origin(self):
        """
        Abstract property to be defined in child class.

        Get the name for the origin of the file. E.g. 'hessio'.

        Returns
        -------
        origin : str
        """

    @staticmethod
    @abstractmethod
    def check_file_compatibility(file_path):
        """
        Abstract method to be defined in child class.

        Perform a set of checks to see if the input file is compatible
        with this file reader.

        Parameters
        ----------
        file_path : str
            File path to the event file.

        Returns
        -------
        compatible : bool
            True if file is compatible, False if it is incompatible
        """

    @property
    @abstractmethod
    def num_events(self):
        """
        Abstract property to be defined in child class.

        Obtain the number of events from the file, store it inside
        self._num_events, and return the value.

        Returns
        -------
        self._num_events : int
        """

    @property
    @abstractmethod
    def event_id_list(self):
        """
        Abstract property to be defined in child class.

        Obtain the number of events from the file, store it as
        self._event_id_list, and return the value.

        Returns
        -------
        self._event_id_list : list[self._num_events]
        """

    @abstractmethod
    def read(self, allowed_tels=None, requested_event=None,
             use_event_id=False):
        """
        Abstract method to be defined in child class.

        Read the file using the processes required by the readers file-type.

        Parameters
        ----------
        allowed_tels : list[int]
            select only a subset of telescope, if None, all are read. This can
            be used for example emulate the final CTA data format, where there
            would be 1 telescope per file (whereas in current monte-carlo,
            they are all interleaved into one file)
        requested_event : int
            Seek to a paricular event index
        use_event_id : bool
            If True ,'requested_event' now seeks for a particular event id
            instead of index

        Returns
        -------
        source : generator
            A generator that can be iterated over to obtain events
        """

    def get_event(self, requested_event, use_event_id=False):
        """
        Loop through events until the requested event is found

        Parameters
        ----------
        requested_event : int
            Seek to a paricular event index
        use_event_id : bool
            If True ,'requested_event' now seeks for a particular event id
            instead of index

        Returns
        -------
        event : `ctapipe` event-container

        """
        source = self.read(requested_event=requested_event,
                           use_event_id=use_event_id)
        event = next(source)
        return deepcopy(event)

    def find_max_true_npe(self, telescopes=None):
        """
        Loop through events to find the maximum true npe

        Parameters
        ----------
        telescopes : list
            List of telecopes to include. If None, then all telescopes
            are included.

        Returns
        -------
        max_pe : int

        """
        # TODO: Find an alternate method so this can be removed
        self.log.info("Finding maximum true npe inside file...")
        source = self.read()
        max_pe = 0
        for event in source:
            tels = list(event.dl0.tels_with_data)
            if telescopes is not None:
                tels = []
                for tel in telescopes:
                    if tel in event.dl0.tels_with_data:
                        tels.append(tel)
            if event.count == 0:
                # Check events have true charge included
                try:
                    if np.all(event.mc.tel[tels[0]].photo_electron_image == 0):
                        raise KeyError
                except KeyError:
                    self.log.exception('[chargeres] Source does not contain '
                                       'true charge')
                    raise
            for telid in tels:
                pe = event.mc.tel[telid].photo_electron_image
                this_max = np.max(pe)
                if this_max > max_pe:
                    max_pe = this_max
        self.log.info("Maximum true npe inside file = {}".format(max_pe))

        return max_pe


class HessioFileReader(EventFileReader):
    name = 'HessioFileReader'
    origin = 'hessio'

    @staticmethod
    def check_file_compatibility(file_path):
        compatible = True
        # TODO: Change check to be a try of hessio_event_source?
        if not file_path.endswith('.gz'):
            compatible = False
        return compatible

    @property
    def num_events(self):
        self.log.info("Obtaining number of events in file...")
        if self._num_events:
            pass
        else:
            self._num_events = len(self.event_id_list)
        self.log.info("Number of events inside file = {}"
                      .format(self._num_events))
        return self._num_events

    @property
    def event_id_list(self):
        self.log.info("Retrieving list of event ids...")
        if self._event_id_list:
            pass
        else:
            self.log.info("Building new list of event ids...")
            ids = hessio_get_list_event_ids(self.input_path,
                                            max_events=self.max_events)
            self._event_id_list = ids
        self.log.info("List of event ids retrieved.")
        return self._event_id_list

    def read(self, allowed_tels=None, requested_event=None,
             use_event_id=False):
        """
        Read the file using the appropriate method depending on the file origin

        Parameters
        ----------
        allowed_tels : list[int]
            select only a subset of telescope, if None, all are read. This can
            be used for example emulate the final CTA data format, where there
            would be 1 telescope per file (whereas in current monte-carlo,
            they are all interleaved into one file)
        requested_event : int
            Seek to a paricular event index
        use_event_id : bool
            If True ,'requested_event' now seeks for a particular event id
            instead of index

        Returns
        -------
        source : generator
            A generator that can be iterated over to obtain events
        """

        # Obtain relevent source
        self.log.debug("Reading file...")
        if self.max_events:
            self.log.info("Max events being read = {}".format(self.max_events))
        source = hessio_event_source(self.input_path,
                                     max_events=self.max_events,
                                     allowed_tels=allowed_tels,
                                     requested_event=requested_event,
                                     use_event_id=use_event_id)
        self.log.debug("File reading complete")
        return source


# External Children
try:
    from targetpipe.io.eventfilereader import TargetioFileReader, \
        ToyioFileReader
except ImportError:
    pass


class EventFileReaderFactory(Factory):
    name = "EventFileReaderFactory"
    description = "Obtain EventFileReader based on file type"

    subclasses = Factory.child_subclasses(EventFileReader)
    subclass_names = [c.__name__ for c in subclasses]

    reader = CaselessStrEnum(subclass_names, None, allow_none=True,
                             help='Event file reader to use. If None then '
                                  'a reader will be chosen based on file '
                                  'extension').tag(config=True)

    # Product classes traits
    # Would be nice to have these automatically set...!
    input_path = Unicode(get_dataset('gamma_test.simtel.gz'), allow_none=True,
                         help='Path to the input file containing '
                              'events.').tag(config=True)
    max_events = Int(None, allow_none=True,
                     help='Maximum number of events that will be read from'
                          'the file').tag(config=True)

    def get_factory_name(self):
        return self.name

    def get_product_name(self):
        if self.reader is not None:
            return self.reader
        else:
            if self.input_path is None:
                raise ValueError("Please specify an input_path for event file")
            try:
                for subclass in self.subclasses:
                    if subclass.check_file_compatibility(self.input_path):
                        return subclass.__name__
                raise ValueError
            except ValueError:
                self.log.exception("Cannot find compatible EventFileReader "
                                   "for: {}".format(self.input_path))
                raise



import pickle
from math import log10, sqrt
from os import makedirs
from os.path import dirname, exists

import numpy as np
from scipy.stats import binned_statistic as bs

from ctapipe.core import Component
from ctapipe.core.traits import Int, Bool

__all__ = ['ChargeResolutionCalculator',]

class ChargeResolutionCalculator(Component):
    """
    Class to handle the calculation of Charge Resolution.

    Attributes
    ----------
    max_pe : int
        Maximum pe to calculate the charge resolution up to.
    sum_dict : dict
        Dictionary to store the running sum for each true charge.
    n_dict : dict
        Dictionary to store the running number for each true charge.
    variation_hist_nbins : float
        Number of bins for the variation histogram.
    variation_hist_range : list
        X and Y range for the variation histogram.
    variation_hist : `np.histogram2d`
    variation_xedges : ndarray
        Edges of the X bins for the variation histogram.
    variation_yedges : ndarray
        Edges of the Y bins for the variation histogram.
    """
    name = 'ChargeResolutionCalculator'

    max_pe = Int(2000, help='Maximum pe to calculate the charge resolution '
                            'up to').tag(config=True)
    binning = Int(60, allow_none=True,
                  help='Number of bins for the Charge Resolution. If None, '
                       'no binning is performed.').tag(config=True)
    log_bins = Bool(True, help='Bin the x axis linearly instead of '
                               'logarithmic.').tag(config=True)

    def __init__(self, config, tool, **kwargs):
        """
        Calculator of charge resolution.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        reductor : ctapipe.calib.camera.reductors.Reductor
            The reductor to use to reduce the waveforms in the event.
            By default no data volume reduction is applied, and the dl0 samples
            will equal the r1 samples.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)

        self.sum_array = np.zeros(self.max_pe)
        self.n_array = np.zeros(self.max_pe)
        self.sum_dict = {}
        self.n_dict = {}

        self.variation_hist_nbins = log10(self.max_pe) * 50
        self.variation_hist_range = [[log10(1), log10(self.max_pe)],
                                     [log10(1), log10(self.max_pe)]]
        h, xedges, yedges = np.histogram2d([np.nan], [np.nan],
                                           bins=self.variation_hist_nbins,
                                           range=self.variation_hist_range)
        self.variation_hist = h
        self.variation_xedges = xedges
        self.variation_yedges = yedges

    def add_charges(self, true_charge, measured_charge):
        """
        Fill the class parameters with a numpy array of true charge and
        measured (calibrated) charge from an event. The two arrays must be the
        same size.

        Parameters
        ----------
        true_charge : ndarray
            Array of true (MC) charge.
            Obtained from event.mc.tel[telid].image[channel]
        measured_charge : ndarray
            Array of measured (dl1 calibrated) charge.
            Obtained from event.mc.tel[tel_id].photo_electron_image
        """
        above_0 = (measured_charge > 0) & (true_charge > 0)
        x = true_charge[above_0]
        y = measured_charge[above_0]
        h, _, _ = np.histogram2d(np.log10(y),
                                 np.log10(x),
                                 bins=self.variation_hist_nbins,
                                 range=self.variation_hist_range)
        self.variation_hist += h

        in_range = (true_charge > 0) & (true_charge <= self.max_pe)
        true_q = true_charge[in_range]
        measured_q = measured_charge[in_range]
        np.add.at(self.sum_array, true_q - 1, np.power(measured_q - true_q, 2))
        np.add.at(self.n_array, true_q - 1, 1)

    def get_charge_resolution(self):
        """
        Calculate and obtain the charge resolution graph arrays.

        Returns
        -------
        true_charge : ndarray
            The X axis true charges.
        chargeres : ndarray
            The Y axis charge resolution values.
        chargeres_error : ndarray
            The error on the charge resolution.
        scaled_chargeres : ndarray
            The Y axis charge resolution divided by the Goal.
        scaled_chargeres_error : ndarray
            The error on the charge resolution divided by the Goal.
        """
        self.log.debug('[chargeres] Calculating charge resolution')

        n_1 = self.n_array > 0
        n = self.n_array[n_1]
        true = (np.arange(self.max_pe) + 1)[n_1]
        sum_ = self.sum_array[n_1]

        res = np.sqrt((sum_ / n) + true) / true
        res_error = res * (1 / np.sqrt(2 * n))

        scale = self.goal(true)
        scaled_res = res / scale
        scaled_res_error = res_error / scale

        if self.binning is not None:
            x = true
            if self.log_bins:
                x = np.log10(true)

            def binning(array):
                return bs(x, array, 'mean', bins=self.binning)

            def sum_errors(array):
                return np.sqrt(np.sum(np.power(array, 2))) / array.size

            def bin_errors(array):
                return bs(x, array, sum_errors, bins=self.binning)

            true, _, _ = binning(true)
            res, _, _ = binning(res)
            res_error, _, _ = bin_errors(res_error)
            scaled_res, _, _ = binning(scaled_res)
            scaled_res_error, _, _ = bin_errors(scaled_res_error)

        return true, res, res_error, scaled_res, scaled_res_error

    @staticmethod
    def limit_curves(npe, n_nsb, n_add, enf, sigma2):
        """
        Equation for calculating the Goal and Requirement curves, as defined
        in SCI-MC/121113.
        https://portal.cta-observatory.org/recordscentre/Records/SCI/
        SCI-MC/measurment_errors_system_performance_1YQCBC.pdf

        Parameters
        ----------
        npe : ndarray
            Number of photoeletrons (variable).
        n_nsb : float
            Number of NSB photons.
        n_add : float
            Number of photoelectrons from additional noise sources.
        enf : float
            Excess noise factor.
        sigma2 : float
            Percentage ofmultiplicative errors.
        """
        return (np.sqrt((n_nsb + n_add) + np.power(enf, 2) * npe +
                        np.power(sigma2 * npe, 2)) / npe).astype(float)

    @staticmethod
    def requirement(npe):
        """
        CTA requirement curve.

        Parameters
        ----------
        npe : ndarray
            Number of photoeletrons (variable).
        """
        n_nsb = sqrt(4.0 + 3.0)
        n_add = 0
        enf = 1.2
        sigma2 = 0.1
        defined_npe = 1000

        # If npe is not an array, temporarily convert it to one
        npe = np.array([npe])

        lc = ChargeResolutionCalculator.limit_curves
        requirement = lc(npe, n_nsb, n_add, enf, sigma2)
        requirement[npe > defined_npe] = np.nan

        return requirement[0]

    @staticmethod
    def goal(npe):
        """
        CTA goal curve.

        Parameters
        ----------
        npe : ndarray
            Number of photoeletrons (variable).
        """
        n_nsb = 2
        n_add = 0
        enf = 1.1152
        sigma2 = 0.05
        defined_npe = 2000

        # If npe is not an array, temporarily convert it to one
        npe = np.array([npe])

        lc = ChargeResolutionCalculator.limit_curves
        goal = lc(npe, n_nsb, n_add, enf, sigma2)
        goal[npe > defined_npe] = np.nan

        return goal[0]

    @staticmethod
    def poisson(npe):
        """
        Poisson limit curve.

        Parameters
        ----------
        npe : ndarray
            Number of photoeletrons (variable).
        """
        # If npe is not an array, temporarily convert it to one
        npe = np.array([npe])
        poisson = np.sqrt(npe) / npe

        return poisson[0]

    def save(self, path):
        d = dict(max_pe=self.max_pe,
                 sum_array=self.sum_array,
                 n_array=self.n_array,
                 variation_hist_nbins=self.variation_hist_nbins,
                 variation_hist_range=self.variation_hist_range,
                 variation_hist=self.variation_hist,
                 variation_xedges=self.variation_xedges,
                 variation_yedges=self.variation_yedges)
        output_dir = dirname(path)
        if not exists(output_dir):
            self.log.info("[output] Creating directory: {}".format(output_dir))
            makedirs(output_dir)
        with open(path, 'wb') as f:
            self.log.info("Saving Charge Resolution file: {}".format(path))
            pickle.dump(d, f, pickle.HIGHEST_PROTOCOL)

    def load(self, path):
        if not path.endswith("pickle"):
            self.log.error("File is not a pickle file: {}".format(path))
        with open(path, 'rb') as f:
            self.log.info("Loading Charge Resolution file: {}".format(path))
            d = pickle.load(f)
            self.max_pe = d['max_pe']
            self.sum_array = d['sum_array']
            self.n_array = d['n_array']
            self.variation_hist_nbins = d['variation_hist_nbins']
            self.variation_hist_range = d['variation_hist_range']
            self.variation_hist = d['variation_hist']
            self.variation_xedges = d['variation_xedges']
            self.variation_yedges = d['variation_yedges']



from copy import deepcopy

import numpy as np

from astropy import units as u
from sklearn.preprocessing import StandardScaler


class RegressorClassifierBase:
    """This class collects one model for every camera type -- given by
    `cam_id_list` -- to get an estimate for the energy of an
    air-shower.  The class interfaces with `scikit-learn` in that it
    relays all function calls not exeplicitly defined here through
    `.__getattr__` to any model in its collection.  This gives us the
    full power and convenience of scikit-learn and we only have to
    bother about the high-level interface.

    The fitting approach is to train on single images; then separately
    predict an image's energy and combine the different predictions
    for all the telscopes in each event in one single energy estimate.

    TODO: come up with a better merging approach to combine the
          estimators of the various camera types

    Parameters
    ----------

    model: scikit-learn model
        the model you want to use to estimate the shower energies
    cam_id_list: list of strings
        list of identifiers to differentiate the various sources of
        the images; could be the camera IDs or even telescope IDs. We
        will train one model for each of the identifiers.
    unit: 1 or astropy unit
        scikit-learn regressors don't work with astropy unit. so, tell
        in advance in which unit we want to deal here in case we need
        one. (default: 1)
    kwargs: **dict
        arguments to be passed on to the constructor of the regressors

    """

    def __init__(self, model, cam_id_list, unit=1, **kwargs):
        self.model_dict = {}
        self.unit = unit
        for cam_id in cam_id_list or []:
            self.model_dict[cam_id] = model(**deepcopy(kwargs))

    def __getattr__(self, attr):
        """We interface this class with the "first" model in `.model_dict`
        and relay all function calls over to it. This gives access to
        all the fancy implementations in the scikit-learn classes
        right from this class and we only have to overwrite the
        high-level interface functions we want to adapt to our
        use-case.

        """
        return getattr(next(iter(self.model_dict.values())), attr)

    def __str__(self):
        """ return the class name of the used model """
        return str(next(iter(self.model_dict.values()))).split("(")[0]

    def reshuffle_event_list(self, X, y):
        """I collect the data event-wise as dictionaries of the different
        camera identifiers but for the training it is more convenient
        to have a dictionary containing flat arrays of the
        feature lists. This function flattens the `X` and `y` arrays
        accordingly.

        This is only for convenience during the training and testing
        phase, later on, we won't collect data first event- and then
        telescope-wise. That's why this is a separate function and not
        implemented in `.fit`.

        Parameters
        ----------
        X : list of dictionaries of lists
            collection of training features see Notes section for a sketch of
            the container's expected shape
        y : list of astropy quantities
            list of the training targets (i.e. shower energies) for all events

        Returns
        -------
        trainFeatures, trainTarget : dictionaries of lists
            flattened containers to be handed over to `.fit`

        Raises
        ------
        KeyError:
            in case `X` contains keys that were not provided with
            `cam_id_list` during `.__init__` or `.load`.

        Notes
        -----
        `X` and `y` are expected to be lists of events:
        `X = [event_1, event_2, ..., event_n]`
        `y = [energy_1, energy_2, ..., energy_n]`

        with `event_i` being a dictionary:
        `event_i = {tel_type_1: list_of_tels_1, ...}`
        and `energy_i` is the energy of `event_i`

        `list_of_tels_i` being the list of all telescope of type
        `tel_type_1` containing the features to fit and predict the
        shower energy: `list_of_tels_i = [[tel_1_feature_1,
        tel_1_feature_2, ...], [tel_2_feature_1, tel_2_feature_2,
        ...], ...]`

        after reshuffling, the resulting lists will look like this::

          trainFeatures = {tel_type_1: [features_event_1_tel_1,
                                       features_event_1_tel_2,
                                       features_event_2_tel_1, ...],
                          tel_type_2: [features_event_1_tel_3,
                                       features_event_2_tel_3,
                                       features_event_2_tel_4, ...],
                          ...}


        `trainTarget` will be a dictionary with the same keys and a
        lists of energies corresponding to the features in the
        `trainFeatures` lists.

        """

        trainFeatures = {a: [] for a in self.model_dict}
        trainTarget = {a: [] for a in self.model_dict}

        for evt, target in zip(X, y):
            for cam_id, tels in evt.items():
                try:
                    # append the features-lists of the current event
                    # and telescope type to the flat list of
                    # features-lists for this telescope type
                    trainFeatures[cam_id] += tels
                except KeyError:
                    raise KeyError("cam_id '{}' in X but no model defined: {}"
                                   .format(cam_id, [k for k in self.model_dict]))

                try:
                    # add a target-entry for every feature-list
                    trainTarget[cam_id] += \
                        [target.to(self.unit).value] * len(tels)
                except AttributeError:
                    # in case the target is not given as an astropy
                    # quantity let's hope that the user keeps proper
                    # track of the unit themself (might be just `1` anyway)
                    trainTarget[cam_id] += [target] * len(tels)
        return trainFeatures, trainTarget

    def fit(self, X, y):
        """This function fits a model against the collected features;
        separately for every telescope identifier.

        Parameters
        ----------
        X : dictionary of lists of lists
            Dictionary that maps the telescope identifiers to lists of
            feature-lists.  The values of the dictionary are the lists
            `scikit-learn` regressors train on and are supposed to
            comply to their format requirements e.g. each featuer-list
            has to contain the same features at the same position
        y : dictionary of lists
            the energies corresponding to all the feature-lists of `X`

        Returns
        -------
        self

        Raises
        ------
        KeyError:
            in case `X` contains keys that are either not in `y` or
            were not provided before with `cam_id_list`.

        """
        for cam_id in X:
            if cam_id not in y:
                raise KeyError("cam_id '{}' in X but not in y: {}"
                               .format(cam_id, [k for k in y]))

            if cam_id not in self.model_dict:
                raise KeyError("cam_id '{}' in X but no model defined: {}"
                               .format(cam_id, [k for k in self.model_dict]))

            # for every `cam_id` train one model (as long as there are events in `X`)
            if len(X[cam_id]):
                self.model_dict[cam_id].fit(X[cam_id], y[cam_id])

        return self

    # def predict(self, X, cam_id=None):
    #     """
    #     In the tradition of scikit-learn, `.predict` takes a "list of feature-lists" and
    #     returns an estimate for targeted quantity for every  set of features.
    #
    #     Parameters
    #     ----------
    #     X : list of lists of floats
    #         the list of feature-lists
    #     cam_id : any (e.g. string, int), optional
    #         identifier of the singular camera type to consider here
    #         if not set, `.reg_dict` is assumed to have a single key which is used in place
    #
    #     Returns
    #     -------
    #     predict : list of floats
    #         predictions for the target quantity for every set of features given in `X`
    #
    #     Raises
    #     ------
    #     ValueError
    #         if `cam_id is None` and the number of registered models is not 1
    #     """
    #
    #     if cam_id is None:
    #         if len(self.model_dict) == 1:
    #             cam_id = next(iter(self.model_dict.keys()))
    #         else:
    #             raise ValueError("you need to provide a cam_id")
    #
    #     return self.model_dict[cam_id].predict(X)*self.energy_unit

    def save(self, path):
        """saves the models in `.reg_dict` each in a separate pickle to disk

        TODO: investigate more stable containers to write out models
        than joblib dumps

        Parameters
        ----------
        path : string
            Path to store the different models.  Expects to contain
            `{cam_id}` or at least an empty `{}` to replace it with
            the keys in `.reg_dict`.

        """

        from sklearn.externals import joblib
        for cam_id, model in self.model_dict.items():
            try:
                # assume that there is a `{cam_id}` keyword to replace
                # in the string
                joblib.dump(model, path.format(cam_id=cam_id))
            except IndexError:
                # if not, assume there is a naked `{}` somewhere left
                # if not, format won't do anything, so it doesn't
                # break but will overwrite every pickle with the
                # following one
                joblib.dump(model, path.format(cam_id))

    @classmethod
    def load(cls, path, cam_id_list, unit=1):
        """Load the pickled dictionary of model from disk, create a husk
        `cls` instance and fill the model dictionary.

        Parameters
        ----------
        path : string
            the path where the pre-trained, pickled regressors are
            stored `path` is assumed to contain a `{cam_id}` keyword
            to be replaced by each camera identifier in `cam_id_list`
            (or at least a naked `{}`).
        cam_id_list : list
            list of camera identifiers like telescope ID or camera ID
            and the assumed distinguishing feature in the filenames of
            the various pickled regressors.
        unit : 1 or astropy unit, optional
            scikit-learn regressor/classifier do not work with
            units. so append this one to the predictions in case you
            deal with unified targets (like energy).  assuming that
            the models where trained with consistent units.  clf
        self : RegressorClassifierBase
            in derived classes, this will return a ready-to-use
            instance of that class to predict any problem you have
            trained for

        """
        from sklearn.externals import joblib

        # need to get an instance of this class `cam_id_list=None`
        # prevents `.__init__` to initialise `.model_dict` itself,
        # since we are going to set it with the pickled models
        # manually
        self = cls(cam_id_list=None, unit=unit)
        for key in cam_id_list:
            try:
                # assume that there is a `{cam_id}` keyword to replace
                # in the string
                self.model_dict[key] = joblib.load(path.format(cam_id=key))
            except IndexError:
                # if not, assume there is a naked `{}` somewhere left
                # if not, format won't do anything, so it doesn't matter
                # though this will load the same model for every `key`
                self.model_dict[key] = joblib.load(path.format(key))

        return self

    @staticmethod
    def scale_features(cam_id_list, feature_list):
        """Scales features before training with any ML method.

        Parameters
        ----------
        cam_id_list : list
            list of camera identifiers like telescope ID or camera ID
            and the assumed distinguishing feature in the filenames of
            the various pickled regressors.
        feature_list : dictionary of lists of lists
            Dictionary that maps the telescope identifiers to lists of
            feature-lists.  The values of the dictionary are the lists
            `scikit-learn` regressors train on and are supposed to
            comply to their format requirements e.g. each feature-list
            has to contain the same features at the same position

        Returns
        -------
        f_dict : dictionary of lists of lists
            a copy of feature_list input, with scaled values

        """
        f_dict = {}
        scaler = {}

        for cam_id in cam_id_list or []:
            scaler[cam_id] = StandardScaler()
            scaler[cam_id].fit(feature_list[cam_id])
            f_dict[cam_id] = scaler[cam_id].transform(feature_list[cam_id])

        return f_dict, scaler

    def show_importances(self, feature_labels=None):
        """Creates a matplotlib figure that shows the importances of the
        different features for the various trained models in a grid of
        barplots.  The features are sorted by descending importance.

        Parameters
        ----------
        feature_labels : list of strings, optional
            a list of the feature names in proper order
            to be used as x-axis tick-labels

        Returns
        -------
        fig : matplotlib figure
            the figure holding the different bar plots

        """
        import matplotlib.pyplot as plt
        n_tel_types = len(self.model_dict)
        n_cols = np.ceil(np.sqrt(n_tel_types)).astype(int)
        n_rows = np.ceil(n_tel_types / n_cols).astype(int)

        fig, axs = plt.subplots(nrows=n_rows, ncols=n_cols, squeeze=False)
        plt.suptitle("Feature Importances")
        for i, (cam_id, model) in enumerate(self.model_dict.items()):
            plt.sca(axs.ravel()[i])
            plt.title(cam_id)
            try:
                importances = model.feature_importances_
            except:
                plt.gca().axis('off')
                continue
            bins = range(importances.shape[0])
            if feature_labels:
                importances, s_feature_labels = \
                    zip(*sorted(zip(importances, feature_labels), reverse=True))
                plt.xticks(bins, s_feature_labels, rotation=17)
            plt.bar(bins, importances,
                    color='r', align='center')

        # switch off superfluous axes
        for j in range(i + 1, n_rows * n_cols):
            axs.ravel()[j].axis('off')

        return fig



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Components to read HESSIO data.  

This requires the hessio python library to be installed
"""
import logging

from astropy import units as u
from astropy.coordinates import Angle
from astropy.time import Time

from .containers import DataContainer
from ..core import Provenance
from ..instrument import TelescopeDescription, SubarrayDescription

logger = logging.getLogger(__name__)

try:
    from pyhessio import open_hessio
    from pyhessio import HessioError
    from pyhessio import HessioTelescopeIndexError
    from pyhessio import HessioGeneralError
except ImportError as err:
    logger.fatal(
        "the `pyhessio` python module is required to access MC data: {}"
            .format(err))
    raise err

__all__ = [
    'hessio_event_source',
]


def hessio_get_list_event_ids(url, max_events=None):
    """
    Faster method to get a list of all the event ids in the hessio file.
    This list can also be used to find out the number of events that exist
    in the file.

    Parameters
    ----------
    url : str
        path to file to open
    max_events : int, optional
        maximum number of events to read

    Returns
    -------
    event_id_list : list[num_events]
        A list with all the event ids that are in the file.

    """
    logger.warning("This method is slow. Need to find faster method.")
    try:
        with open_hessio(url) as pyhessio:
            Provenance().add_input_file(url, role='r0.sub.evt')
            counter = 0
            event_id_list = []
            eventstream = pyhessio.move_to_next_event()
            for event_id in eventstream:
                event_id_list.append(event_id)
                counter += 1
                if max_events and counter >= max_events:
                    pyhessio.close_file()
                    break
            return event_id_list
    except HessioError:
        raise RuntimeError("hessio_event_source failed to open '{}'"
                           .format(url))


def hessio_event_source(url, max_events=None, allowed_tels=None,
                        requested_event=None, use_event_id=False):
    """A generator that streams data from an EventIO/HESSIO MC data file
    (e.g. a standard CTA data file.)

    Parameters
    ----------
    url : str
        path to file to open
    max_events : int, optional
        maximum number of events to read
    allowed_tels : list[int]
        select only a subset of telescope, if None, all are read. This can
        be used for example emulate the final CTA data format, where there
        would be 1 telescope per file (whereas in current monte-carlo,
        they are all interleaved into one file)
    requested_event : int
        Seek to a paricular event index
    use_event_id : bool
        If True ,'requested_event' now seeks for a particular event id instead
        of index
    """

    with open_hessio(url) as pyhessio:
        # the container is initialized once, and data is replaced within
        # it after each yield
        Provenance().add_input_file(url, role='dl0.sub.evt')
        counter = 0
        eventstream = pyhessio.move_to_next_event()
        if allowed_tels is not None:
            allowed_tels = set(allowed_tels)
        data = DataContainer()
        data.meta['origin'] = "hessio"

        # some hessio_event_source specific parameters
        data.meta['input'] = url
        data.meta['max_events'] = max_events

        for event_id in eventstream:

            # Seek to requested event
            if requested_event is not None:
                current = counter
                if use_event_id:
                    current = event_id
                if not current == requested_event:
                    counter += 1
                    continue

            data.r0.run_id = pyhessio.get_run_number()
            data.r0.event_id = event_id
            data.r0.tels_with_data = set(pyhessio.get_teldata_list())
            data.r1.run_id = pyhessio.get_run_number()
            data.r1.event_id = event_id
            data.r1.tels_with_data = set(pyhessio.get_teldata_list())
            data.dl0.run_id = pyhessio.get_run_number()
            data.dl0.event_id = event_id
            data.dl0.tels_with_data = set(pyhessio.get_teldata_list())

            # handle telescope filtering by taking the intersection of
            # tels_with_data and allowed_tels
            if allowed_tels is not None:
                selected = data.r0.tels_with_data & allowed_tels
                if len(selected) == 0:
                    continue  # skip event
                data.r0.tels_with_data = selected
                data.r1.tels_with_data = selected
                data.dl0.tels_with_data = selected

            data.trig.tels_with_trigger \
                = pyhessio.get_central_event_teltrg_list()
            time_s, time_ns = pyhessio.get_central_event_gps_time()
            data.trig.gps_time = Time(time_s * u.s, time_ns * u.ns,
                                      format='unix', scale='utc')
            data.mc.energy = pyhessio.get_mc_shower_energy() * u.TeV
            data.mc.alt = Angle(pyhessio.get_mc_shower_altitude(), u.rad)
            data.mc.az = Angle(pyhessio.get_mc_shower_azimuth(), u.rad)
            data.mc.core_x = pyhessio.get_mc_event_xcore() * u.m
            data.mc.core_y = pyhessio.get_mc_event_ycore() * u.m
            first_int = pyhessio.get_mc_shower_h_first_int() * u.m
            data.mc.h_first_int = first_int

            # mc run header data
            data.mcheader.run_array_direction = \
                pyhessio.get_mc_run_array_direction()

            data.count = counter

            # this should be done in a nicer way to not re-allocate the
            # data each time (right now it's just deleted and garbage
            # collected)

            data.r0.tel.clear()
            data.r1.tel.clear()
            data.dl0.tel.clear()
            data.dl1.tel.clear()
            data.mc.tel.clear()  # clear the previous telescopes

            _fill_instrument_info(data, pyhessio)

            for tel_id in data.r0.tels_with_data:

                # event.mc.tel[tel_id] = MCCameraContainer()

                data.mc.tel[tel_id].dc_to_pe \
                    = pyhessio.get_calibration(tel_id)
                data.mc.tel[tel_id].pedestal \
                    = pyhessio.get_pedestal(tel_id)

                data.r0.tel[tel_id].adc_samples = \
                    pyhessio.get_adc_sample(tel_id)
                data.r0.tel[tel_id].adc_sums = \
                    pyhessio.get_adc_sum(tel_id)
                data.mc.tel[tel_id].reference_pulse_shape = \
                    pyhessio.get_ref_shapes(tel_id)

                nsamples = pyhessio.get_event_num_samples(tel_id)
                if nsamples <= 0:
                    nsamples = 1
                data.r0.tel[tel_id].num_samples = nsamples

                # load the data per telescope/pixel
                hessio_mc_npe = pyhessio.get_mc_number_photon_electron
                data.mc.tel[tel_id].photo_electron_image \
                    = hessio_mc_npe(telescope_id=tel_id)
                data.mc.tel[tel_id].meta['refstep'] = \
                    pyhessio.get_ref_step(tel_id)
                data.mc.tel[tel_id].time_slice = \
                    pyhessio.get_time_slice(tel_id)
                data.mc.tel[tel_id].azimuth_raw = \
                    pyhessio.get_azimuth_raw(tel_id)
                data.mc.tel[tel_id].altitude_raw = \
                    pyhessio.get_altitude_raw(tel_id)
                data.mc.tel[tel_id].azimuth_cor = \
                    pyhessio.get_azimuth_cor(tel_id)
                data.mc.tel[tel_id].altitude_cor = \
                    pyhessio.get_altitude_cor(tel_id)
            yield data
            counter += 1

            if max_events and counter >= max_events:
                pyhessio.close_file()
                return


def _fill_instrument_info(data, pyhessio):
    """
    fill the data.inst structure with instrumental information.

    Parameters
    ----------
    data: DataContainer
        data container to fill in

    """
    if not data.inst.telescope_ids:
        data.inst.telescope_ids = list(pyhessio.get_telescope_ids())
        data.inst.subarray = SubarrayDescription("MonteCarloArray")

        for tel_id in data.inst.telescope_ids:
            try:

                pix_pos = pyhessio.get_pixel_position(tel_id) * u.m
                foclen = pyhessio.get_optical_foclen(tel_id) * u.m
                mirror_area = pyhessio.get_mirror_area(tel_id) * u.m ** 2
                num_tiles = pyhessio.get_mirror_number(tel_id)
                tel_pos = pyhessio.get_telescope_position(tel_id) * u.m

                tel = TelescopeDescription.guess(*pix_pos, foclen)
                tel.optics.mirror_area = mirror_area
                tel.optics.num_mirror_tiles = num_tiles
                data.inst.subarray.tels[tel_id] = tel
                data.inst.subarray.positions[tel_id] = tel_pos

                # deprecated fields that will become part of
                # TelescopeDescription or SubrrayDescription
                data.inst.optical_foclen[tel_id] = foclen
                data.inst.pixel_pos[tel_id] = pix_pos
                data.inst.tel_pos[tel_id] = tel_pos

                nchans = pyhessio.get_num_channel(tel_id)
                npix = pyhessio.get_num_pixels(tel_id)
                data.inst.num_channels[tel_id] = nchans
                data.inst.num_pixels[tel_id] = npix
                data.inst.mirror_dish_area[tel_id] = mirror_area
                data.inst.mirror_numtiles[tel_id] = num_tiles

            except HessioGeneralError:
                pass
